"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=too-many-lines
from abc import abstractmethod
from functools import reduce
from enum import Enum
import typing
from typing import Dict, List, Tuple
from copy import deepcopy

import numpy as np

from loom.eka.utilities import ndarray_rowsum, is_tableau_valid
from ..operations.base_operation import Operation
from ..operations.gate_operation import (
    Identity,
    Hadamard,
    Phase,
    PhaseInv,
    X,
    Z,
    Y,
    CNOT,
    CY,
    CZ,
    SWAP,
)
from ..operations.measurement_operation import Measurement, Reset
from ..operations.resize_operation import AddQubit, DeleteQubit
from ..tableau import Tableau
from ..operations.datamanipulation_operation import (
    UpdateTableau,
    CreatePauliFrame,
    RecordPauliFrame,
    CreateClassicalRegister,
    RecordClassicalRegister,
)
from ..operations.classical_operation import (
    ClassicalNOT,
    ClassicalOR,
    ClassicalAND,
)
from ..operations.controlled_operation import ControlledOperation
from ..pauli_frame import PauliFrame
from ..data_store import DataStore
from ..classicalreg import ClassicalRegister
from ..exceptions import (
    InvalidTableauError,
    ClassicalRegisterError,
    ClassicalOperationError,
)


class Instruction:
    """An Instruction contains the implementation logic for a quantum operation.

    Every Instruction requires an `apply` method for Tableau-based simulations,
    and a `apply_pf` and `apply_pf_back` method for PauliFrame-based simulations.

    PauliFrame simulations can be performed by propagating a PauliFrame forward or
    backward through a circuit. The methods `apply_pf` and `apply_pf_back` are used to
    propagate the PauliFrame forward and backward respectively.

    The Instruction class is an abstract class and should not be instantiated directly.

    Raises
    ------
    NotImplementedError
        If the method is not implemented in the subclass.
    """

    @abstractmethod
    def apply(self, input_tableau: Tableau):
        """
        Abstract method that applies the quantum operation to the input Tableau.
        """
        return NotImplementedError

    # pylint: disable=unused-argument
    def apply_pf(self, input_pauliframes: List[PauliFrame]):
        """
        Abstract method that propagates the PauliFrames forward through the quantum
        operation.
        """
        return NotImplementedError

    def apply_pf_back(self, input_pauliframes: List[PauliFrame]):
        """
        Abstract method that propagates the PauliFrames backward through the quantum
        operation.
        """
        return NotImplementedError


class IdentityInstruction(Instruction):
    """
    A simple Instruction that does nothing. This will be the initial Instruction in
    a Moment. Other Instructions within the Moment will decorate this Instruction.
    """

    # pylint: disable=arguments-differ
    def apply(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[Tableau, DataStore, Dict]:
        """
        Applies the quantum operation to the input Tableau.
        """
        # print("Applying I")
        return input_tableau, data_store, kwargs

    def apply_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[List[PauliFrame], DataStore, Dict]:
        """
        Propagates the PauliFrames forward through the quantum operation.
        """
        return input_pauliframes, data_store, kwargs

    def apply_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[List[PauliFrame], DataStore, Dict]:
        """
        Propagates the PauliFrames backward through the quantum operation.
        """
        return input_pauliframes, data_store, kwargs


class InstructionDecorator(Instruction):
    """An InstructionDecorator is a wrapper around an Instruction that adds additional
    functionality to the Instruction. The wrapped Instruction is then used to apply
    quantum operations to a Tableau, or PauliFrame, representation of a quantum state.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.base_operation.Operation`
        The Operation that will be applied.

    Attributes
    ----------
    wrapped_instruction : Instruction
        The Instruction that is being decorated.
    input_operation : :class:`loom.cliffordsim.operations.base_operation.Operation`
        The quantum operation that is being applied to the quantum state.
    """

    def __init__(self, instruction: Instruction, input_operation: Operation):
        self.wrapped_instruction = instruction
        if self.__class__.__name__[:-9] != input_operation.__class__.__name__:
            raise TypeError(
                "This InstructionDecorator only accepts                 "
                f" {self.__class__.__name__[:-9]} Operation type."
            )
        self.input_operation = input_operation

    # pylint: disable=arguments-differ
    def apply(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[Tableau, DataStore, Dict]:
        """
        Applies the quantum operation to the input Tableau.
        """
        input_tableau, data_store, kwargs = self.extra(
            input_tableau, data_store, **kwargs
        )
        return self.wrapped_instruction.apply(input_tableau, data_store, **kwargs)

    @abstractmethod
    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[Tableau, DataStore, Dict]:
        """
        An abstract method that adds additional functionality to the `apply` method.
        """
        return NotImplementedError

    def apply_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[List[PauliFrame], DataStore, Dict]:
        """
        Applies the quantum operation to the input PauliFrames.
        """
        input_pauliframes, data_store, kwargs = self.extra_pf(
            input_pauliframes, data_store, **kwargs
        )
        return self.wrapped_instruction.apply_pf(
            input_pauliframes, data_store, **kwargs
        )

    @abstractmethod
    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[List[PauliFrame], DataStore, Dict]:
        """
        An abstract method that adds additional functionality to the `apply_pf` method.
        """
        return NotImplementedError

    def apply_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[List[PauliFrame], DataStore, Dict]:
        """
        Applies the quantum operation to the input PauliFrames. (For backward
        propagation.)
        """
        input_pauliframes, data_store, kwargs = self.extra_pf_back(
            input_pauliframes, data_store, **kwargs
        )
        return self.wrapped_instruction.apply_pf_back(
            input_pauliframes, data_store, **kwargs
        )

    @abstractmethod
    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ) -> Tuple[List[PauliFrame], DataStore, Dict]:
        """
        An abstract method that adds additional functionality to the `apply_pf_back`
        method.
        """
        return NotImplementedError


class IdentityDecorator(InstructionDecorator):
    """The IdentityDecorator returns the input Tableau or PauliFrames without any
    modifications.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.Identity`
        The Identity Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to apply the Identity gate to.
    """

    def __init__(self, instruction: Instruction, input_operation: Identity):
        super().__init__(instruction, input_operation)

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class HadamardDecorator(InstructionDecorator):
    """The HadamardDecorator applies a Hadamard gate to a qubit in a quantum state.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.Hadamard`
        The Hadamard Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to apply the Hadamard gate to.
    """

    def __init__(self, instruction: Instruction, input_operation: Hadamard):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        self.apply_hadamard(input_tableau, self.qubit)

        return input_tableau, data_store, kwargs

    @staticmethod
    def apply_hadamard(input_tableau: Tableau, qubit: int):
        """Applies a hadamard gate on a specific qubit.

        Parameters
        ----------
        input_tableau : Tableau
            The input representation of the quantum state.
        qubit : int
            The qubit to apply the Hadamard gate to.
        """
        # print("Applying H")
        input_tableau.r ^= input_tableau.x[:, qubit] & input_tableau.z[:, qubit]
        # swap x and z without using temp
        input_tableau.x[:, qubit] ^= input_tableau.z[:, qubit]
        input_tableau.z[:, qubit] ^= input_tableau.x[:, qubit]
        input_tableau.x[:, qubit] ^= input_tableau.z[:, qubit]

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        input_pauliframes = self.apply_hadamard_pf(input_pauliframes, self.qubit)
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        The Hadamard gate is its own inverse, so applying it again will revert the
        PauliFrame to its original state.
        """
        input_pauliframes = self.apply_hadamard_pf(input_pauliframes, self.qubit)
        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_hadamard_pf(
        input_pauliframes: List[PauliFrame], qubit: int
    ) -> List[PauliFrame]:
        """
        Applies a Hadamard gate on a specific qubit in a list of PauliFrames.
        """
        # swap z and x values
        for input_pauliframe in input_pauliframes:
            input_pauliframe.z[qubit], input_pauliframe.x[qubit] = (
                input_pauliframe.x[qubit],
                input_pauliframe.z[qubit],
            )
        return input_pauliframes


class PhaseDecorator(InstructionDecorator):
    """The PhaseDecorator applies a phase gate, Z(+pi/2), to a qubit in a quantum state.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.Phase`
        The Phase Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to apply the phase gate to.
    """

    def __init__(self, instruction: Instruction, input_operation: Phase):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        self.apply_phase(input_tableau, self.qubit)
        return input_tableau, data_store, kwargs

    @staticmethod
    def apply_phase(input_tableau: Tableau, qubit: int):
        """Applies a phase gate on a specific qubit.

        Parameters
        ----------
        input_tableau : Tableau
            The input representation of the quantum state.
        qubit : int
            The qubit to apply the phase gate to.
        """
        input_tableau.r ^= input_tableau.x[:, qubit] & input_tableau.z[:, qubit]
        input_tableau.z[:, qubit] ^= input_tableau.x[:, qubit]

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        input_pauliframes = self.apply_phase_pf(input_pauliframes, self.qubit)

        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        The inverse of the phase gate is the phase inverse gate, which can be applied
        using the same transformation as the phase gate in the PauliFrame
        representation.
        """
        input_pauliframes = self.apply_phase_pf(input_pauliframes, self.qubit)

        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_phase_pf(
        input_pauliframes: List[PauliFrame], qubit: int
    ) -> List[PauliFrame]:
        """
        Applies a phase gate on a specific qubit in a list of PauliFrames.
        """
        # Z<->Z, X<->Y
        for input_pauliframe in input_pauliframes:
            input_pauliframe.z[qubit] ^= input_pauliframe.x[qubit]
        return input_pauliframes


class PhaseInvDecorator(InstructionDecorator):
    """The PhaseInvDecorator applies the inverse of a phase gate, Z(-pi/2), to a qubit
    in a quantum state.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.PhaseInv`
        The PhaseInv Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to apply the inverse of the phase gate to.
    """

    def __init__(self, instruction: Instruction, input_operation: PhaseInv):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit

    @staticmethod
    def apply_phase_inv(input_tableau: Tableau, qubit: int):
        """Applies the inverse of a phase gate on a specific qubit.

        Parameters
        ----------
        input_tableau : Tableau
            The input representation of the quantum state.
        qubit : int
            The qubit to apply the inverse of the phase gate to.
        """
        # X-> -Y -> -X -> Y -> X
        input_tableau.r ^= input_tableau.x[:, qubit] & (~input_tableau.z[:, qubit])
        input_tableau.z[:, qubit] ^= input_tableau.x[:, qubit]

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        self.apply_phase_inv(input_tableau, self.qubit)
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        input_pauliframes = PhaseDecorator.apply_phase_pf(input_pauliframes, self.qubit)

        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        input_pauliframes = PhaseDecorator.apply_phase_pf(input_pauliframes, self.qubit)

        return input_pauliframes, data_store, kwargs


class XDecorator(InstructionDecorator):
    """The XDecorator applies an X gate to a qubit in a quantum state.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.X`
        The X Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to apply the X gate to.
    """

    def __init__(self, instruction: Instruction, input_operation: X):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # print("Applying X")
        self.apply_x(input_tableau, self.qubit)
        return input_tableau, data_store, kwargs

    @staticmethod
    def apply_x(input_tableau: Tableau, qubit: int):
        """Applies an X gate on a specific qubit.

        Parameters
        ----------
        input_tableau : Tableau
            The input representation of the quantum state.
        qubit : int
            The qubit to apply the X gate to.
        """
        input_tableau.r ^= input_tableau.z[:, qubit]

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class ZDecorator(InstructionDecorator):
    """The ZDecorator applies a Z gate to a qubit in a quantum state.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.Z`
        The Z Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to apply the Z gate to.
    """

    def __init__(self, instruction: Instruction, input_operation: Z):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # print("Applying Z")
        input_tableau.r ^= input_tableau.x[:, self.qubit]
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class YDecorator(InstructionDecorator):
    """The YDecorator applies a Y gate to a qubit in a quantum state.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.Y`
        The Y Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to apply the Y gate to.
    """

    def __init__(self, instruction: Instruction, input_operation: Y):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # print("Applying Y")
        input_tableau.r ^= (
            input_tableau.x[:, self.qubit] ^ input_tableau.z[:, self.qubit]
        )
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class CNOTDecorator(InstructionDecorator):
    """The CNOTDecorator applies a CNOT, Controlled-NOT, gate across 2 qubits, a
    control qubit and a target qubit.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.CNOT`
        The CNOT Operation that will be applied.

    Attributes
    ----------
    c_qubit : int
        The control qubit that controls the CNOT gate.
    t_qubit : int
        The target qubit where the NOT/X gate will be applied if the control qubit is 1.
    """

    def __init__(self, instruction: Instruction, input_operation: CNOT):
        super().__init__(instruction, input_operation)
        self.c_qubit = input_operation.control_qubit
        self.t_qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # print("Applying CNOT")
        input_tableau.r ^= (
            input_tableau.x[:, self.c_qubit]
            & input_tableau.z[:, self.t_qubit]
            & ~(input_tableau.x[:, self.t_qubit] ^ input_tableau.z[:, self.c_qubit])
        )

        input_tableau.x[:, self.t_qubit] ^= input_tableau.x[:, self.c_qubit]
        input_tableau.z[:, self.c_qubit] ^= input_tableau.z[:, self.t_qubit]
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        Applies a CNOT gate to a list of PauliFrames.
        """
        input_pauliframes = self.apply_cnot_pf(
            input_pauliframes, self.c_qubit, self.t_qubit
        )

        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_cnot_pf(
        input_pauliframes: List[PauliFrame], c_qubit: int, t_qubit: int
    ) -> List[PauliFrame]:
        """
        A staticmethod that applies a CNOT gate to a list of PauliFrames.
        """
        for input_pauliframe in input_pauliframes:
            # XI -> XX
            input_pauliframe.x[t_qubit] ^= input_pauliframe.x[c_qubit]
            # IZ -> ZZ
            input_pauliframe.z[c_qubit] ^= input_pauliframe.z[t_qubit]

        return input_pauliframes

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        input_pauliframes = self.apply_cnot_pf(
            input_pauliframes, self.c_qubit, self.t_qubit
        )

        return input_pauliframes, data_store, kwargs


class CYDecorator(InstructionDecorator):
    """The CYDecorator applies a CY, Controlled-Y, gate across 2 qubits, a control qubit
    and a target qubit.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.CY`
        The CY Operation that will be applied.

    Attributes
    ----------
    c_qubit : int
        The control qubit that controls the CY gate.
    t_qubit : int
        The target qubit where the Y gate will be applied if the control qubit is 1.
    """

    def __init__(self, instruction: Instruction, input_operation: CY):
        super().__init__(instruction, input_operation)
        self.c_qubit = input_operation.control_qubit
        self.t_qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        input_tableau.r ^= (
            input_tableau.x[:, self.c_qubit]
            & (input_tableau.x[:, self.t_qubit] ^ input_tableau.z[:, self.t_qubit])
            & ~(input_tableau.z[:, self.c_qubit] ^ input_tableau.z[:, self.t_qubit])
        )

        # IX->ZX, IZ->ZZ but IY->IY
        input_tableau.z[:, self.c_qubit] ^= (
            input_tableau.x[:, self.t_qubit] ^ input_tableau.z[:, self.t_qubit]
        )

        # XI->XY
        input_tableau.x[:, self.t_qubit] ^= input_tableau.x[:, self.c_qubit]
        input_tableau.z[:, self.t_qubit] ^= input_tableau.x[:, self.c_qubit]
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        Applies a CY gate to a list of PauliFrames.
        """
        input_pauliframes = self.apply_cy_pf(
            input_pauliframes, self.c_qubit, self.t_qubit
        )
        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_cy_pf(
        input_pauliframes: List[PauliFrame], c_qubit: int, t_qubit: int
    ) -> List[PauliFrame]:
        """
        A staticmethod that applies a CY gate to a list of PauliFrames.
        """
        for input_pauliframe in input_pauliframes:
            input_pauliframe.z[c_qubit] ^= (
                input_pauliframe.x[t_qubit] ^ input_pauliframe.z[t_qubit]
            )
            input_pauliframe.x[t_qubit] ^= input_pauliframe.x[c_qubit]
            input_pauliframe.z[t_qubit] ^= input_pauliframe.x[c_qubit]
        return input_pauliframes

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        self.apply_cy_pf(input_pauliframes, self.c_qubit, self.t_qubit)

        return input_pauliframes, data_store, kwargs


class CZDecorator(InstructionDecorator):
    """The CZDecorator applies a CZ, Controlled-Z, gate across 2 qubits, a control qubit
    and a target qubit.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.CZ`
        The CZ Operation that will be applied.

    Attributes
    ----------
    c_qubit : int
        The control qubit that controls the CZ gate.
    t_qubit : int
        The target qubit where the Z gate will be applied if the control qubit is 1.
    """

    def __init__(self, instruction: Instruction, input_operation: CZ):
        super().__init__(instruction, input_operation)
        self.c_qubit = input_operation.control_qubit
        self.t_qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # print("Applying CZ")
        input_tableau.r ^= (
            input_tableau.x[:, self.c_qubit]
            & input_tableau.x[:, self.t_qubit]
            & ~(input_tableau.z[:, self.t_qubit] ^ input_tableau.z[:, self.c_qubit])
        )

        input_tableau.z[:, self.t_qubit] ^= input_tableau.x[:, self.c_qubit]
        input_tableau.z[:, self.c_qubit] ^= input_tableau.x[:, self.t_qubit]
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        Applies a CZ gate to a list of PauliFrames.
        """
        input_pauliframes = self.apply_cz_pf(
            input_pauliframes, self.c_qubit, self.t_qubit
        )

        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_cz_pf(
        input_pauliframes: List[PauliFrame], c_qubit: int, t_qubit: int
    ) -> List[PauliFrame]:
        """
        A staticmethod that applies a CZ gate to a list of PauliFrames.
        """
        for input_pauliframe in input_pauliframes:
            input_pauliframe.z[t_qubit] ^= input_pauliframe.x[c_qubit]
            input_pauliframe.z[c_qubit] ^= input_pauliframe.x[t_qubit]
        return input_pauliframes

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        self.apply_cz_pf(input_pauliframes, self.c_qubit, self.t_qubit)

        return input_pauliframes, data_store, kwargs


class SWAPDecorator(InstructionDecorator):
    """The SWAPDecorator applies a SWAP gate across 2 qubits, labelled as control
    qubit and target qubit.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.SWAP`
        The SWAP Operation that will be applied.

    Attributes
    ----------
    c_qubit : int
        The qubit that will have its state "swapped" with the target qubit.
    t_qubit : int
        The qubit that will have its state "swapped" with the control qubit.
    """

    def __init__(self, instruction: Instruction, input_operation: SWAP):
        super().__init__(instruction, input_operation)
        self.c_qubit = input_operation.control_qubit
        self.t_qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # print("Applying SWAP")

        # swap the columns of the 2 qubits for x and z arrays
        input_tableau.x[:, [self.t_qubit, self.c_qubit]] = input_tableau.x[
            :, [self.c_qubit, self.t_qubit]
        ]
        input_tableau.z[:, [self.t_qubit, self.c_qubit]] = input_tableau.z[
            :, [self.c_qubit, self.t_qubit]
        ]

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        Applies a SWAP gate to a list of PauliFrames.
        """
        self.apply_swap_pf(input_pauliframes, self.c_qubit, self.t_qubit)

        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_swap_pf(
        input_pauliframes: List[PauliFrame], c_qubit: int, t_qubit: int
    ) -> List[PauliFrame]:
        """
        A staticmethod that applies a SWAP gate to a list of PauliFrames.
        """
        for input_pauliframe in input_pauliframes:
            # swap Z and X
            input_pauliframe.z[c_qubit], input_pauliframe.z[t_qubit] = (
                input_pauliframe.z[t_qubit],
                input_pauliframe.z[c_qubit],
            )

            input_pauliframe.x[c_qubit], input_pauliframe.x[t_qubit] = (
                input_pauliframe.x[t_qubit],
                input_pauliframe.x[c_qubit],
            )

        return input_pauliframes

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        Applies a SWAP gate to a list of PauliFrames. (For backward propagation.)
        The SWAP gate is its own inverse, so applying it again will revert the
        PauliFrame to its original state.
        """
        input_pauliframes = self.apply_swap_pf(
            input_pauliframes, self.c_qubit, self.t_qubit
        )

        return input_pauliframes, data_store, kwargs


class ClassicalBitDecorator(InstructionDecorator):  # pylint: disable=abstract-method
    """
    The ClassicalBitDecorator contains private methods to get information about a bit
    in a classical register, these methods will be used by its subclasses.

    Any InstructionDecorator that involves classical bits should inherit from this
    class.
    """

    def _get_bit_info(self, classical_register, bit_order, bit_id):
        """
        Get information of a bit in the classical register defined either by the
        ordering within the classical register or its bit ID.
        """
        bit_value = None
        if bit_order is not None:
            bit_value, bit_order, bit_id = self._get_bit_info_w_bit_order(
                classical_register, bit_order
            )
        elif bit_id is not None:
            bit_value, bit_order, bit_id = self._get_bit_info_w_bit_id(
                classical_register, bit_id
            )

        return bit_value, bit_order, bit_id

    def _get_bit_info_w_bit_order(self, classical_register, bit_order):
        """Get information of a bit in the classical register defined by its ordering
        within the classical register.
        """
        try:
            bit_id, bit_value = classical_register.reg[bit_order]

            return bit_value, bit_order, bit_id

        except Exception as exc:
            raise ClassicalOperationError(
                f"The selected bit_order, {bit_order}, does not exist or is not valid "
                f"in {classical_register.name}. There are only "
                f"{len(classical_register.reg)} bits in the register."
            ) from exc

    def _get_bit_info_w_bit_id(self, classical_register, bit_id):
        """Get information of a bit in the classical register defined by its bit ID."""
        try:
            bit_value = classical_register.id_bit_reg[bit_id]
            for each_order, each_pair in enumerate(classical_register.reg):
                if each_pair[0] == bit_id:
                    bit_order = each_order
                    break
            return bit_value, bit_order, bit_id

        except Exception as exc:
            raise ClassicalOperationError(
                f"The selected bit_id, {bit_id}, does not exist or is not valid in "
                f"{classical_register.name}. The only available bit IDs in this "
                f"register are: {classical_register.bit_ids}"
            ) from exc

    def _select_register(self, input_registry: dict, register_name: str):
        """Get a classical register from the input registry with the given name."""
        try:
            select_register = input_registry[register_name]

            return select_register

        except Exception as exc:
            raise ClassicalRegisterError(
                "An error has occured when trying to select the "
                f"register, {register_name}."
            ) from exc


class MeasurementDecorator(ClassicalBitDecorator):
    """The MeasurementDecorator applies a measurement operation to a qubit in a quantum
    state. It also records the measurement result in a classical register if it exists.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.MeasurementOperation`
        The measurement operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to be measured.
    measurement_id : str
        The label of the measurement operation.
    bias : float
        The bias of the measurement result.
    reg_name : str
        The name of the classical register where the measurement result will be stored.
    bit_order : int
        The order of the bit in the classical register.
    bit_id : int
        The ID of the bit in the classical register.
    """

    def __init__(self, instruction: Instruction, input_operation: Measurement):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit
        self.measurement_id = input_operation.label
        self.bias = input_operation.bias
        self.basis = input_operation.basis
        self.reg_name = input_operation.reg_name
        self.bit_order = input_operation.bit_order
        self.bit_id = input_operation.bit_id

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        m_qubit = self.qubit

        # Apply any necessary operations to the tableau before measurement to change
        # the basis of the measurement.
        if self.basis == "X":
            HadamardDecorator.apply_hadamard(input_tableau, m_qubit)
        elif self.basis == "Y":
            # To map |+i> to |0> and |-i> to |1>, we need to a PhaseInv and a Hadamard
            # gate.
            PhaseInvDecorator.apply_phase_inv(input_tableau, m_qubit)
            HadamardDecorator.apply_hadamard(input_tableau, m_qubit)
        elif self.basis == "Z":
            pass
        else:
            raise ValueError(
                f"Invalid measurement basis: {self.basis}. Must be 'X', 'Y', or 'Z'."
            )

        # Find pivot index
        is_result_random, measurement_result = self.measure(
            input_tableau, m_qubit, self.bias
        )

        # Add measured results to DataStore
        data_store.record_measurements(
            self.measurement_id, measurement_result, is_result_random
        )

        # Write to a Classical Register if it exists
        if self.reg_name is not None:
            output_reg = self._select_register(kwargs["registry"], self.reg_name)
            bit_value, bit_order, bit_id = self._get_bit_info(
                output_reg, self.bit_order, self.bit_id
            )

            # Write output to register
            bit_value = measurement_result
            kwargs["registry"][self.reg_name].reg[bit_order] = (bit_id, bit_value)

        return input_tableau, data_store, kwargs

    @classmethod
    def measure(
        cls, input_tableau: Tableau, m_qubit: int, bias: float = 0.5
    ) -> tuple[bool, int]:
        """
        Determines the measurement result of a qubit and whether it is random.
        It also updates the input tableau if the measurement result is random.

        Parameters
        ----------
        input_tableau : Tableau
            The input tableau of the quantum state.
        m_qubit : int
            The qubit to be measured.
        bias : float, optional
            The bias of the result, by default 0.5

        Returns
        -------
        tuple[bool, int]
            A tuple containing whether the measurement result is random and what its
            value is.
        """
        p = cls.find_pivot(input_tableau, m_qubit)
        # Find if m result shall be random
        is_result_random = cls.is_result_random(input_tableau, m_qubit, p)
        # Deduce measurement result
        measurement_result = (
            cls.deduce_mresult_probabilistic(input_tableau, m_qubit, p, bias)
            if is_result_random
            else cls.deduce_mresult_deterministic(input_tableau, m_qubit)
        )

        return is_result_random, measurement_result

    @staticmethod
    def find_pivot(input_tableau: Tableau, m_qubit: int) -> int:
        """
        Finds the pivot point for the measurement operation.

        Parameters
        ----------
        input_tableau : Tableau
            The input tableau of the quantum state.
        m_qubit : int
            Qubit to be measured

        Returns
        -------
        int
            Pivot index
        """
        # nqubits is added since it was sliced out of the search
        return (
            np.argmax(input_tableau.x[input_tableau.nqubits :, m_qubit] == 1)
            + input_tableau.nqubits
        )

    @staticmethod
    def is_result_random(input_tableau: Tableau, m_qubit: int, p: int) -> bool:
        """
        Checks if the measurement of a qubit, will result in a
        random outcome. The pivot point has to also be given.

        Parameters
        ----------
        input_tableau : Tableau
            The input tableau of the quantum state.
        m_qubit : int
            Qubit to be measured
        p : int
            Pivot index

        Returns
        -------
        bool
            Whether the result will be random or not.
        """
        return input_tableau.x[p, m_qubit] == 1

    @staticmethod
    def deduce_mresult_deterministic(
        input_tableau: Tableau,
        m_qubit: int,
    ) -> typing.Literal[0, 1]:
        """
        Performs measurement operation and returns the result.

        Parameters
        ----------
        input_tableau : Tableau
            The input tableau of the quantum state.
        m_qubit : int
            Qubit to be measured

        Returns
        -------
        typing.Literal[0, 1]
            Result of the measurement operation.
        """

        nqubits = input_tableau.nqubits
        # Set scratch row to all zeros
        input_tableau.scratch_row[:] = np.zeros_like(input_tableau.scratch_row)
        # Apply rowsum appropriately to get result in scratch row
        input_tableau.tableau_w_scratch[:, :] = reduce(
            lambda tab, idx: (
                ndarray_rowsum(tab, -1, idx + nqubits)
                if input_tableau.x[idx, m_qubit] == 1
                else tab
            ),
            range(nqubits),
            input_tableau.tableau_w_scratch,
        )

        # obtain measurement result
        measurement_result = input_tableau.scratch_row[-1]
        return measurement_result

    @staticmethod
    def deduce_mresult_probabilistic(
        input_tableau: Tableau, m_qubit: int, p: int, bias: float = 0.5
    ) -> typing.Literal[0, 1]:
        """
        Performs measurement operation and returns the result for a random
        measurement.

        Parameters
        ----------
        input_tableau : Tableau
            The input tableau of the quantum state.
        m_qubit : int
            Qubit to be measured
        p : int
            Pivot index
        bias : float, optional
            The bias of the result, by default 0.5
            If this is set to 0 then the result will always be 0.

        Returns
        -------
        typing.Literal[0, 1]
            Result of the measurement operation.
        """
        nqubits = input_tableau.nqubits
        # Firstly, apply rowsum on the appropriate indices
        idcs_wo_p = np.concatenate(
            [
                np.arange(p - nqubits),
                np.arange(p - nqubits + 1, p),
                np.arange(p + 1, 2 * nqubits),
            ]
        )
        # Apply rowsum only if x[idx, m_qubit] == 1
        input_tableau.tableau_w_scratch[:, :] = reduce(
            lambda tab, idx: (
                ndarray_rowsum(tab, idx, p)
                if input_tableau.x[idx, m_qubit] == 1
                else tab
            ),
            idcs_wo_p,
            input_tableau.tableau_w_scratch,
        )

        # Secondly modify line p-n
        input_tableau.tableau[p - input_tableau.nqubits, :] = input_tableau.tableau[
            p, :
        ]

        # Third modify line p, to make it into ±III...Z...I
        input_tableau.tableau[p, :] = np.zeros_like(input_tableau.tableau[p, :])
        input_tableau.z[p, m_qubit] = 1
        # Obtain measurement result
        measurement_result = 0
        if input_tableau.rand_gen.random() < bias:
            # r has be set to 0 already, so flip it
            input_tableau.r[p] = 1
            measurement_result = 1
        return measurement_result

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        input_pauliframes, flip_results = self.apply_meas_pf(
            input_pauliframes, self.qubit, self.basis
        )

        data_store.record_measurement_from_pauliframes(
            self.measurement_id, flip_results, input_pauliframes
        )

        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_meas_pf(
        input_pauliframes: List[PauliFrame], qubit: int, basis: str
    ) -> tuple[List[PauliFrame], List[int]]:
        """
        Applies a measurement operation to a list of PauliFrames and returns the
        modified PauliFrames along with the measurement results.
        """

        flip_results = []
        if basis == "Z":
            flip_results = [
                1 if input_pauliframe.x[qubit] == 1 else 0
                for input_pauliframe in input_pauliframes
            ]
        elif basis == "X":
            flip_results = [
                1 if input_pauliframe.z[qubit] == 1 else 0
                for input_pauliframe in input_pauliframes
            ]
        elif basis == "Y":
            flip_results = [
                1 if (input_pauliframe.x[qubit] ^ input_pauliframe.z[qubit]) == 1 else 0
                for input_pauliframe in input_pauliframes
            ]

        return input_pauliframes, flip_results

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        Note that changes that causes the pauliframes to interact with Datastore
        when propagating forward, are not properly implemented here for the backward
        pass. Basically ignore the problem until we have a better solution.
        """
        input_pauliframes, _ = self.apply_meas_pf(
            input_pauliframes, self.qubit, self.basis
        )

        return input_pauliframes, data_store, kwargs


class AddQubitDecorator(InstructionDecorator):
    """The AddQubitDecorator adds a qubit to the quantum state. If the index exists in
    the current quantum state, the new qubit will be inserted at the index while
    the existing qubits with an index larger than the inserted qubits index
    will have their index increased by 1.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: AddQubit
        The AddQubit Operation that will be applied.

    Attributes
    ----------
    index : int
        The index of the qubit to be added.
    """

    def __init__(self, instruction: Instruction, input_operation: AddQubit):
        super().__init__(instruction, input_operation)
        self.index = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        prev_nqubits = input_tableau.nqubits
        new_nqubits = prev_nqubits + 1
        # add columns
        zero_column = np.zeros(2 * prev_nqubits + 1, dtype=np.int8).reshape(
            2 * prev_nqubits + 1, 1
        )

        tab = input_tableau.tableau_w_scratch

        input_tableau.tableau_w_scratch = np.concatenate(
            [
                tab[:, : self.index],
                zero_column,
                tab[:, self.index : prev_nqubits + self.index],
                zero_column,
                tab[:, prev_nqubits + self.index :],
            ],
            axis=1,
        )

        # add rows

        zero_row = np.zeros(2 * new_nqubits + 1, dtype=np.int8).reshape(
            1, 2 * new_nqubits + 1
        )

        tab = input_tableau.tableau_w_scratch
        input_tableau.tableau_w_scratch = np.concatenate(
            [
                tab[: self.index, :],
                zero_row,
                tab[self.index : prev_nqubits + self.index, :],
                zero_row,
                tab[prev_nqubits + self.index :, :],
            ],
            axis=0,
        )

        # set the destabilizer of this qubit to X
        input_tableau.tableau_w_scratch[self.index, self.index] = 1
        # set the stabilizer of this qubit to Z
        input_tableau.tableau_w_scratch[
            new_nqubits + self.index, new_nqubits + self.index
        ] = 1

        input_tableau.update()

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        input_pauliframes = self.apply_add_qubit_pf(input_pauliframes, self.index)
        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_add_qubit_pf(
        input_pauliframes: List[PauliFrame], index: int
    ) -> List[PauliFrame]:
        """
        Applies an AddQubit operation to a list of PauliFrames.
        A staticmethod that applies an AddQubit operation to a list of PauliFrames.
        """
        for input_pauliframe in input_pauliframes:
            # add 0 to the new index
            input_pauliframe.z = np.insert(input_pauliframe.z, index, 0)
            input_pauliframe.x = np.insert(input_pauliframe.x, index, 0)
        return input_pauliframes

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        # remove the qubit at the index since it's backwards Addition
        input_pauliframes = DeleteQubitDecorator.apply_delete_qubit_pf(
            input_pauliframes, self.index
        )

        return input_pauliframes, data_store, kwargs


class DeleteQubitDecorator(InstructionDecorator):
    """The DeleteQubitDecorator deletes a qubit from the quantum state. If the index
    exists in the current quantum state, the qubit will be removed from the state
    while the existing qubits with an index larger than the deleted qubits index
    will have their index decreased by 1.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: DeleteQubit
        The DeleteQubit Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to be deleted.
    """

    def __init__(self, instruction: Instruction, input_operation: DeleteQubit):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # find pivot
        p = MeasurementDecorator.find_pivot(input_tableau, self.qubit)

        # find if result shall be random
        is_result_random = MeasurementDecorator.is_result_random(
            input_tableau, self.qubit, p
        )

        if not is_result_random:
            # make it random by applying an H gate
            HadamardDecorator.apply_hadamard(input_tableau, self.qubit)

            # find pivot again
            p = MeasurementDecorator.find_pivot(input_tableau, self.qubit)

            # find if result shall be random
            is_result_random = MeasurementDecorator.is_result_random(
                input_tableau, self.qubit, p
            )

            assert is_result_random, "The measurement result should be random."

        # We bias the measurement to come out positive such that we can safely delete
        # the row containing the +Z_i operator.
        # If we allow -Z_i as a result, any instance of other operator containing the
        # Z_i operator should have its sign flipped when the row is deleted.
        MeasurementDecorator.deduce_mresult_probabilistic(
            input_tableau, self.qubit, p, bias=0
        )

        # find which rowws to delete
        nqubits = input_tableau.nqubits
        stab_row_idx = p
        destab_row_idx = p - nqubits
        # delete the rows
        input_tableau.tableau_w_scratch = np.delete(
            input_tableau.tableau_w_scratch,
            [stab_row_idx, destab_row_idx],
            axis=0,
        )

        # delete the columns of the
        z_col = self.qubit + nqubits
        x_col = self.qubit
        input_tableau.tableau_w_scratch = np.delete(
            input_tableau.tableau_w_scratch,
            [z_col, x_col],
            axis=1,
        )

        input_tableau.update()

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        Applies a DeleteQubit operation to a list of PauliFrames.
        """
        input_pauliframes = self.apply_delete_qubit_pf(input_pauliframes, self.qubit)

        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_delete_qubit_pf(
        input_pauliframes: List[PauliFrame], index: int
    ) -> List[PauliFrame]:
        """
        A staticmethod that applies a DeleteQubit operation to a list of PauliFrames.
        """
        for input_pauliframe in input_pauliframes:
            # Z and X entries are deleted
            input_pauliframe.z = np.delete(input_pauliframe.z, index)
            input_pauliframe.x = np.delete(input_pauliframe.x, index)
        return input_pauliframes

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        # Backwards deletion is addition
        input_pauliframes = AddQubitDecorator.apply_add_qubit_pf(
            input_pauliframes, self.qubit
        )
        return input_pauliframes, data_store, kwargs


class ResetDecorator(InstructionDecorator):
    """The ResetDecorator applies a reset operation to a qubit in a quantum state. The
    qubit is first measured, if the measurement result of the qubit is 1, an X gate
    will be applied to the qubit to reset it to 0.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: :class:`loom.cliffordsim.operations.Reset`
        The Reset Operation that will be applied.

    Attributes
    ----------
    qubit : int
        The qubit to be reset.
    """

    def __init__(self, instruction: Instruction, input_operation: Reset):
        super().__init__(instruction, input_operation)
        self.qubit = input_operation.target_qubit
        self.state = input_operation.state

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # First measure the qubit
        _, mres = MeasurementDecorator.measure(input_tableau, self.qubit)

        # Figure out what corrections are needed
        mres_needed = 0
        apply_h = False
        apply_s = False
        match self.state:
            case "0":
                pass
            case "1":
                mres_needed = 1
            case "+":
                apply_h = True
            case "-":
                mres_needed, apply_h = 1, True
            case "+i":
                apply_h, apply_s = True, True
            case "-i":
                mres_needed, apply_h, apply_s = 1, True, True
            case _:
                raise RuntimeError("Incorrect state given for reset operation.")

        # Get to the state that we need to be in
        if mres != mres_needed:
            XDecorator.apply_x(input_tableau, self.qubit)
        if apply_h:
            HadamardDecorator.apply_hadamard(input_tableau, self.qubit)
        if apply_s:
            PhaseDecorator.apply_phase(input_tableau, self.qubit)

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        """
        Applies a Reset operation to a list of PauliFrames.
        """
        input_pauliframes = self.apply_reset_pf(input_pauliframes, self.qubit)

        return input_pauliframes, data_store, kwargs

    @staticmethod
    def apply_reset_pf(
        input_pauliframes: List[PauliFrame], qubit: int
    ) -> List[PauliFrame]:
        """
        A staticmethod that applies a Reset operation to a list of PauliFrames
        """
        for input_pauliframe in input_pauliframes:
            # Z and X errors disappear
            input_pauliframe.z[qubit] = 0
            input_pauliframe.x[qubit] = 0

        return input_pauliframes

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        input_pauliframes = self.apply_reset_pf(input_pauliframes, self.qubit)

        return input_pauliframes, data_store, kwargs


class UpdateTableauDecorator(InstructionDecorator):
    """
    The UpdateTableauDecorator updates the tableau of the quantum state with a new
    tableau. The new tableau must be a numpy ndarray. The decorator can also validate
    the tableau before updating.

    The new tableau must be of the same shape as the current tableau. (Ignoring the
    scratch row)

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: UpdateTableau
        The UpdateTableau Operation that will be applied.

    Attributes
    ----------
    tableau : np.ndarray
        The new tableau to be updated.
    validate : bool
        A flag to validate the tableau before updating.
    """

    def __init__(self, instruction: Instruction, input_operation: UpdateTableau):
        super().__init__(instruction, input_operation)
        self.tableau = input_operation.tableau
        self.validate = input_operation.validate

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        if not isinstance(self.tableau, np.ndarray):
            raise InvalidTableauError("The given tableau must be of type numpy ndarray")

        if self.validate:
            if not is_tableau_valid(self.tableau):
                raise InvalidTableauError("The given tableau is not valid.")

        # Update Tableau here
        input_tableau = input_tableau.rewrite_tableau(self.tableau)

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class CreatePauliFrameDecorator(InstructionDecorator):
    """
    The CreatePauliFrameDecorator creates a new PauliFrame object and adds it to the
    list of propagating PauliFrames at runtime.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: CreatePauliFrame
        The CreatePauliFrame Operation that will be applied.

    Attributes
    ----------
    pauli_frame : PauliFrame
        The PauliFrame object to be created.
    """

    def __init__(self, instruction: Instruction, input_operation: CreatePauliFrame):
        super().__init__(instruction, input_operation)
        self.pauli_frame = deepcopy(input_operation.pauli_frame)

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        # Add the PauliFrame to the list of propagating frames
        # only if the direction of propagation is forward
        if self.pauli_frame.direction == "forward":
            input_pauliframes.append(self.pauli_frame)
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        # Add the PauliFrame to the list of propagating frames
        # only if the direction of propagation is backward
        if self.pauli_frame.direction == "backward":
            input_pauliframes.append(self.pauli_frame)
        return input_pauliframes, data_store, kwargs


class RecordPauliFrameDecorator(InstructionDecorator):
    """
    The RecordPauliFrameDecorator records the state of a PauliFrame object in the
    DataStore. The PauliFrame object must be in the list of propagating PauliFrames at
    runtime.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: RecordPauliFrame
        The RecordPauliFrame Operation that will be applied.

    Attributes
    ----------
    pauli_frame : PauliFrame
        The PauliFrame object to be recorded.
    """

    def __init__(self, instruction: Instruction, input_operation: RecordPauliFrame):
        super().__init__(instruction, input_operation)
        self.pauli_frame = deepcopy(input_operation.pauli_frame)

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        if self.pauli_frame.direction == "forward":
            for input_pauliframe in input_pauliframes:
                if input_pauliframe.id == self.pauli_frame.id:
                    # Record the state of pauli_frame in DataStore
                    data_store.record_pauli_frame(self.pauli_frame, input_pauliframe)
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        if self.pauli_frame.direction == "backward":
            for input_pauliframe in input_pauliframes:
                if input_pauliframe.id == self.pauli_frame.id:
                    # Record the state of pauli_frame in DataStore
                    data_store.record_pauli_frame(self.pauli_frame, input_pauliframe)
        return input_pauliframes, data_store, kwargs


class CreateClassicalRegisterDecorator(InstructionDecorator):
    """
    The CreateClassicalRegisterDecorator creates a new classical register and adds it
    to the registry in the Engine at runtime.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: CreateClassicalRegister
        The CreateClassicalRegister Operation that will be applied.

    Attributes
    ----------
    reg_name : str
        The name of the classical register.
    no_of_bits : int
        The number of bits in the classical register.
    bit_ids : List[int]
        The IDs of the bits in the classical register.
    """

    def __init__(
        self, instruction: Instruction, input_operation: CreateClassicalRegister
    ):
        super().__init__(instruction, input_operation)
        self.reg_name = input_operation.reg_name
        self.no_of_bits = input_operation.no_of_bits
        self.bit_ids = input_operation.bit_ids

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        resolve_creg_names = list(kwargs["registry"].keys())
        if self.reg_name in resolve_creg_names:
            raise ClassicalRegisterError(
                f"The Classical Register of the same name, {self.reg_name}, already "
                "exists."
            )
        kwargs["registry"].update(
            {
                self.reg_name: ClassicalRegister(
                    self.reg_name, self.no_of_bits, self.bit_ids
                )
            }
        )

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class RecordClassicalRegisterDecorator(InstructionDecorator):
    """
    The RecordClassicalRegisterDecorator records the state of a classical register in
    the DataStore. The classical register must exist in the registry in the Engine at
    runtime.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: RecordClassicalRegister
        The RecordClassicalRegister Operation that will be applied.

    Attributes
    ----------
    reg_name : str
        The name of the classical register.
    """

    def __init__(
        self, instruction: Instruction, input_operation: RecordClassicalRegister
    ):
        super().__init__(instruction, input_operation)
        self.reg_name = input_operation.reg_name

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # Record the current state of classical register named reg_name into the
        # DataStore.
        data_store.record_classical_register(kwargs["registry"][self.reg_name])

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class ClassicalNOTDecorator(ClassicalBitDecorator):
    """
    The ClassicalNOTDecorator applies a NOT operation to a bit in a classical
    register.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: ClassicalNOT
        The ClassicalNOT Operation that will be applied.

    Attributes
    ----------
    reg_name : str
        The name of the classical register.
    bit_order : int
        The order of the bit in the classical register.
    bit_id : int
        The ID of the bit in the classical register.
    """

    def __init__(self, instruction: Instruction, input_operation: ClassicalNOT):
        super().__init__(instruction, input_operation)
        self.reg_name = input_operation.reg_name
        self.bit_order = input_operation.bit_order
        self.bit_id = input_operation.bit_id

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # Select by Order or ID. Get info on the bit.
        input_reg = self._select_register(kwargs["registry"], self.reg_name)

        bit_value, bit_order, bit_id = self._get_bit_info(
            input_reg, self.bit_order, self.bit_id
        )

        # Flip the bit
        bit_value ^= 1
        kwargs["registry"][self.reg_name].reg[bit_order] = (bit_id, bit_value)

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class ClassicalORDecorator(ClassicalBitDecorator):
    """
    The ClassicalORDecorator applies a bitwise OR operation to two bits in a classical
    register and stores the result in another bit in the same register. (If the output
    bit exists, it will be overwritten)

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: ClassicalOR
        The ClassicalOR Operation that will be applied.

    Attributes
    ----------
    reg_name : str
        The name of the classical register.
    input_bit_order : List[int]
        The order of the input bits in the classical register.
    input_bit_ids : List[int]
        The IDs of the input bits in the classical register.
    output_reg_name : str
        The name of the classical register where the output bit is stored.
    write_bit_order : int
        The order of the output bit in the classical register.
    write_bit_id : int
        The ID of the output bit in the classical register.
    """

    def __init__(self, instruction: Instruction, input_operation: ClassicalOR):
        super().__init__(instruction, input_operation)
        self.reg_name = input_operation.reg_name
        self.input_bit_order = input_operation.input_bit_order
        self.input_bit_ids = input_operation.input_bit_ids
        self.output_reg_name = input_operation.output_reg_name
        self.write_bit_order = input_operation.write_bit_order
        self.write_bit_id = input_operation.write_bit_id

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # Get Value of Input Bits by Order or ID.
        input_bit_1, input_bit_2 = None, None
        input_reg = self._select_register(kwargs["registry"], self.reg_name)
        if self.input_bit_order != []:
            input_bit_1, _, _ = self._get_bit_info_w_bit_order(
                input_reg, self.input_bit_order[0]
            )
            input_bit_2, _, _ = self._get_bit_info_w_bit_order(
                input_reg, self.input_bit_order[1]
            )
        elif self.input_bit_ids != []:
            input_bit_1, _, _ = self._get_bit_info_w_bit_id(
                input_reg, self.input_bit_ids[0]
            )
            input_bit_2, _, _ = self._get_bit_info_w_bit_id(
                input_reg, self.input_bit_ids[1]
            )

        # Get Information on the Output Bit
        output_reg = self._select_register(kwargs["registry"], self.output_reg_name)
        _, write_bit_order, write_bit_id = self._get_bit_info(
            output_reg, self.write_bit_order, self.write_bit_id
        )

        # Evaluate Operation and Write to Output Bit
        kwargs["registry"][self.output_reg_name].reg[write_bit_order] = (
            write_bit_id,
            input_bit_1 | input_bit_2,
        )

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class ClassicalANDDecorator(ClassicalBitDecorator):
    """
    The ClassicalANDDecorator applies a bitwise AND operation to two bits in a
    classical register and stores the result in another bit in the same register. (If
    the output bit exists, it will be overwritten)

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: ClassicalAND
        The ClassicalAND Operation that will be applied.

    Attributes
    ----------
    reg_name : str
        The name of the classical register.
    input_bit_order : List[int]
        The order of the input bits in the classical register.
    input_bit_ids : List[int]
        The IDs of the input bits in the classical register.
    output_reg_name : str
        The name of the classical register where the output bit is stored.
    write_bit_order : int
        The order of the output bit in the classical register.
    write_bit_id : int
        The ID of the output bit in the classical register.
    """

    def __init__(self, instruction: Instruction, input_operation: ClassicalAND):
        super().__init__(instruction, input_operation)
        self.reg_name = input_operation.reg_name
        self.input_bit_order = input_operation.input_bit_order
        self.input_bit_ids = input_operation.input_bit_ids
        self.output_reg_name = input_operation.output_reg_name
        self.write_bit_order = input_operation.write_bit_order
        self.write_bit_id = input_operation.write_bit_id

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # Get Value of Input Bits by Order or ID.
        input_bit_1, input_bit_2 = None, None
        input_reg = self._select_register(kwargs["registry"], self.reg_name)
        if self.input_bit_order != []:
            input_bit_1, _, _ = self._get_bit_info_w_bit_order(
                input_reg, self.input_bit_order[0]
            )
            input_bit_2, _, _ = self._get_bit_info_w_bit_order(
                input_reg, self.input_bit_order[1]
            )
        elif self.input_bit_ids != []:
            input_bit_1, _, _ = self._get_bit_info_w_bit_id(
                input_reg, self.input_bit_ids[0]
            )
            input_bit_2, _, _ = self._get_bit_info_w_bit_id(
                input_reg, self.input_bit_ids[1]
            )

        # Get Information on the Output Bit
        output_reg = self._select_register(kwargs["registry"], self.output_reg_name)
        _, write_bit_order, write_bit_id = self._get_bit_info(
            output_reg, self.write_bit_order, self.write_bit_id
        )

        # Evaluate Operation and Write to Output Bit
        kwargs["registry"][self.output_reg_name].reg[write_bit_order] = (
            write_bit_id,
            input_bit_1 & input_bit_2,
        )

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


class ControlledOperationDecorator(InstructionDecorator):
    """
    The ControlledOperationDecorator allows the user to transform an Operation into
    a Controlled Operation conditioned on a bit in the classical register. The operation
    is applied if the control bit is 1.

    Parameters
    ----------
    instruction: Instruction
        The instruction that is being decorated.
    input_operation: ControlledOperation
        The ControlledOperation Operation that will be applied. The information about
        the conditioned classical bit and the operation to be applied is stored in the
        `app_operation` of the ControlledOperation.

    Attributes
    ----------
    app_operation : :class:`loom.cliffordsim.operations.base_operation.Operation`
        The operation to be applied.
    reg_name : str
        The name of the classical register.
    bit_order : int
        The order of the bit in the classical register.
    bit_id : int
        The ID of the bit in the classical register.
    """

    def __init__(self, instruction: Instruction, input_operation: ControlledOperation):
        super().__init__(instruction, input_operation)
        self.app_operation = input_operation.app_operation
        self.reg_name = input_operation.reg_name
        self.bit_order = input_operation.bit_order
        self.bit_id = input_operation.bit_id

    @property
    def _app_instruction(self) -> Instruction:
        """Creates the controlled Instruction from the Operation."""
        return getattr(DecoratorSelector, self.app_operation.name).value(
            IdentityInstruction(), self.app_operation
        )

    def extra(
        self,
        input_tableau: Tableau,
        data_store: DataStore,
        **kwargs,
    ):
        # Select by Order or ID. Get info on the bit.
        bit_value = None
        if self.bit_order is not None:
            _, bit_value = kwargs["registry"][self.reg_name].reg[self.bit_order]
        elif self.bit_id is not None:
            for _, each_bit_info in enumerate(kwargs["registry"][self.reg_name].reg):
                if each_bit_info[0] == self.bit_id:
                    bit_value = each_bit_info[1]

        # Apply the Instruction if the bit_value is 1.
        if bit_value == 1:
            input_tableau, data_store, kwargs = self._app_instruction.apply(
                input_tableau, data_store, **kwargs
            )

        return input_tableau, data_store, kwargs

    def extra_pf(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs

    def extra_pf_back(
        self,
        input_pauliframes: List[PauliFrame],
        data_store: DataStore,
        **kwargs,
    ):
        return input_pauliframes, data_store, kwargs


# pylint: disable=invalid-name
class DecoratorSelector(Enum):
    """
    The Decorator Selector chooses which Decorator to call based on the name
    of the input Operation. Decorators should not be selected directly other
    than through this selector.
    """

    DeleteQubit = DeleteQubitDecorator
    AddQubit = AddQubitDecorator
    SWAP = SWAPDecorator
    Identity = IdentityDecorator
    Hadamard = HadamardDecorator
    Phase = PhaseDecorator
    PhaseInv = PhaseInvDecorator
    CNOT = CNOTDecorator
    CY = CYDecorator
    CZ = CZDecorator
    X = XDecorator
    Z = ZDecorator
    Y = YDecorator
    Measurement = MeasurementDecorator
    Reset = ResetDecorator
    UpdateTableau = UpdateTableauDecorator
    CreatePauliFrame = CreatePauliFrameDecorator
    RecordPauliFrame = RecordPauliFrameDecorator
    CreateClassicalRegister = CreateClassicalRegisterDecorator
    RecordClassicalRegister = RecordClassicalRegisterDecorator
    ClassicalNOT = ClassicalNOTDecorator
    ClassicalOR = ClassicalORDecorator
    ClassicalAND = ClassicalANDDecorator
    ControlledOperation = ControlledOperationDecorator



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit, Stabilizer, PauliOperator
from loom.eka.utilities import (
    Direction,
    Orientation,
)
from loom.interpreter import InterpretationStep, Cbit

from loom_rotated_surface_code.code_factory import RotatedSurfaceCode

from .y_wall_out_utilities import qubit_mapper


def y_wall_out_measurement_and_hadamard(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    wall_position: int,
    wall_orientation: Orientation,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Measure a wall of qubits in the Y basis and apply Hadamard gates to the
    appropriate qubits to implement the y_wall_out operation. This is the core of
    a logical phase operation in the rotated surface code.

    - A.) Geometry

        - A.1) Find geometry-dependent variables
        - A.2) Find the qubits to measure, to idle, and to Hadamard

    - B.) Circuit Generation

        - B.1) Generate the circuit of the operation
        - B.2) Append circuit

    - C.) Stabilizer evolution and updates

        - C.1) Identify idle stabilizers that remain unchanged
        - C.2) Identify Hadamard stabilizers that change and create new ones
        - C.3) Reorder data qubits of new Hadamard stabilizers to match original \
            stabilizers order of data qubits per pauli type
        - C.4) Identify wall stabilizers that get merged and create new ones
        - C.5) Define stabilizer evolution and append to interpretation step
        - C.6) Define stabilizer updates and append to interpretation step

    - D.) Logical operator evolution and updates

        - D.1) Find new X logical operator and stabilizers for logical operator jump
        - D.2) Define logical X evolution
        - D.3) Define logical X updates
        
    - E.) Define new Block and update Block history and evolution


    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step.
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    wall_position : int
        The position of the wall.
    wall_orientation : Orientation
        The orientation of the wall (horizontal or vertical).
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        The updated interpretation step.
    """

    # A.) Geometry
    # A.1) Find geometry-dependent variables
    is_wall_hor = wall_orientation == Orientation.HORIZONTAL
    is_top_left_bulk_stab_x = block.upper_left_4body_stabilizer.pauli[0] == "X"
    # A.2) Find the qubits to measure, to idle, and to Hadamard
    qubits_to_measure, qubits_to_idle, qubits_to_hadamard = find_qubit_sets(
        block, wall_position, is_wall_hor
    )

    # B.) Circuit Generation
    # B.1) Generate the circuit of the operation
    circuit, y_meas_cbits = measure_y_and_hadamard_circuit(
        block, qubits_to_measure, qubits_to_hadamard, interpretation_step
    )
    # B.2) Append circuit
    interpretation_step.append_circuit_MUT(
        circuit,
        same_timeslice=same_timeslice,
    )

    # C.) Stabilizer evolution and updates
    new_stabilizers = find_new_stabilizers_and_define_evolution_and_updates(
        interpretation_step,
        block,
        qubits_to_measure,
        qubits_to_idle,
        qubits_to_hadamard,
        y_meas_cbits,
    )

    # D.) Logical operator evolution and updates
    # D.1) Find new X logical operator and stabilizers for logical operator jump
    new_x_log_op, stabilizers_for_x_operator_jump = find_new_x_logical_operator(
        block,
        is_top_left_bulk_stab_x,
        qubits_to_idle,
    )
    # D.2) Define logical X evolution
    interpretation_step.logical_x_evolution.update(
        {
            new_x_log_op.uuid: (block.logical_x_operators[0].uuid,),
        }
    )
    # D.3) Define logical X updates
    interpretation_step.update_logical_operator_updates_MUT(
        "X",
        new_x_log_op.uuid,
        # - Add the z operator updates (because it's a phase gate)
        interpretation_step.logical_z_operator_updates.get(
            block.logical_z_operators[0].uuid, ()
        )
        +
        # - Add the stabilizer cbits needed to redefine the logical operator
        interpretation_step.retrieve_cbits_from_stabilizers(
            stabilizers_for_x_operator_jump, block
        )
        +
        # - Add the y measurement cbits
        tuple(y_meas_cbits) +
        # - Change the parity if the number of measured qubits is 5, 9, 13, ...
        # That is because of the accumulation of the imaginary units from the Y
        # measurements
        (1 * (len(y_meas_cbits) % 4 == 1),),
        inherit_updates=True,
    )

    # E.) Define new Block and update Block history and evolution
    block_without_y_wall_qubits = RotatedSurfaceCode(
        stabilizers=new_stabilizers,
        logical_x_operators=(new_x_log_op,),
        logical_z_operators=block.logical_z_operators,
        unique_label=block.unique_label,
        skip_validation=not debug_mode,
    )
    interpretation_step.update_block_history_and_evolution_MUT(
        (block_without_y_wall_qubits,), (block,)
    )

    return interpretation_step


def find_new_stabilizers_and_define_evolution_and_updates(
    # pylint: disable=too-many-locals
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    qubits_to_measure: list[tuple[int, ...]],
    qubits_to_idle: list[tuple[int, ...]],
    qubits_to_hadamard: list[tuple[int, ...]],
    y_meas_cbits: list[Cbit],
) -> list[Stabilizer]:
    """
    Find the new stabilizers after the Y wall measurement and Hadamard operation.
    Also define the stabilizer evolution and updates in the interpretation step.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step. Note that it may be mutated by generating new
        channels.
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    qubits_to_measure : list[tuple[int, ...]]
        The list of qubits to measure.
    qubits_to_idle : list[tuple[int, ...]]
        The list of qubits to idle.
    qubits_to_hadamard : list[tuple[int, ...]]
        The list of qubits to apply Hadamard gates to.
    y_meas_cbits : list[Cbit]
        The classical bits corresponding to the Y measurements.

    Returns
    -------
    list[Stabilizer]
        The new stabilizers after the Y wall measurement and Hadamard operation.
    """

    # Find vector from idle to hadamard side
    upper_left_idle_qubit = min(qubits_to_idle, key=lambda q: q[0] + q[1])
    upper_left_hadamard_qubit = min(qubits_to_hadamard, key=lambda q: q[0] + q[1])
    idle_to_had_dir = Direction.from_vector(
        (
            upper_left_hadamard_qubit[0] - upper_left_idle_qubit[0],
            upper_left_hadamard_qubit[1] - upper_left_idle_qubit[1],
            0,
        )
    )

    # C.1) Identify idle stabilizers that remain unchanged
    old_idle_stabilizers = [
        stab
        for stab in block.stabilizers
        if set(stab.data_qubits).issubset(qubits_to_idle)
    ]
    new_idle_stabilizers = old_idle_stabilizers

    # C.2) Identify Hadamard stabilizers that change and create new ones
    old_had_stabilizers = [
        stab
        for stab in block.stabilizers
        if set(stab.data_qubits).issubset(qubits_to_hadamard)
    ]
    new_had_stabilizers_with_unordered_data_qubits = [
        Stabilizer(
            pauli="".join({"X": "Z", "Z": "X"}[p] for p in stab.pauli),
            data_qubits=stab.data_qubits,
            ancilla_qubits=stab.ancilla_qubits,
        )
        for stab in old_had_stabilizers
    ]
    # C.3) Reorder data qubits of new Hadamard stabilizers to match original stabilizers
    # order of data qubits per pauli type
    # Take all BULK stabilizers and put their data qubits in the same order as the
    # original stabilizers that were in the position same of the initial block.
    # This is needed because this affects the syndrome extraction circuit.
    new_had_stabilizers = [
        Stabilizer(
            data_qubits=(
                next(
                    [qubit_mapper(q, idle_to_had_dir) for q in init_stab.data_qubits]
                    for init_stab in block.stabilizers
                    if set(
                        (
                            qubit_mapper(q, idle_to_had_dir)
                            for q in init_stab.data_qubits
                        )
                    )
                    == set(stab.data_qubits)
                )
                if len(stab.data_qubits) == 4
                else stab.data_qubits
            ),
            ancilla_qubits=stab.ancilla_qubits,
            pauli=stab.pauli,
        )
        for stab in new_had_stabilizers_with_unordered_data_qubits
    ]

    # C.4) Identify wall stabilizers that get merged and create new ones
    # Wall stabilizers get merged
    old_wall_stabilizers_idle = [
        stab
        for stab in block.stabilizers
        if set(stab.data_qubits).intersection(qubits_to_measure)
        and set(stab.data_qubits).intersection(qubits_to_idle)
        and len(stab.data_qubits) == 4
    ]
    old_wall_stabilizers_had = [
        stab
        for stab in block.stabilizers
        if set(stab.data_qubits).intersection(qubits_to_measure)
        and set(stab.data_qubits).intersection(qubits_to_hadamard)
        and len(stab.data_qubits) == 4
    ]
    # At this point, we need to order the hadamard wall stabilizers so that they
    # correspond to the idle wall stabilizers
    old_wall_stabilizers_had_ordered = [
        old_h_stab
        for old_i_stab in old_wall_stabilizers_idle
        for old_h_stab in old_wall_stabilizers_had
        if len(set(old_h_stab.data_qubits).intersection(old_i_stab.data_qubits)) == 2
    ]

    # Create new wall stabilizers by merging the idle and hadamard wall stabilizers
    new_wall_stabilizers = []
    for old_i_stab in old_wall_stabilizers_idle:
        new_stab_qubits = [
            (
                q
                if q not in qubits_to_measure
                else (
                    q[0] + idle_to_had_dir.to_vector()[0],
                    q[1] + idle_to_had_dir.to_vector()[1],
                    q[2],
                )
            )
            for q in old_i_stab.data_qubits
        ]
        new_stab = Stabilizer(
            old_i_stab.pauli,
            new_stab_qubits,
            ancilla_qubits=old_i_stab.ancilla_qubits,
        )
        new_wall_stabilizers.append(new_stab)

    # C.5) Define stabilizer evolution and append to interpretation step
    stab_evolution = {}
    stab_evolution |= {
        new_stab.uuid: (old_stab.uuid,)
        for old_stab, new_stab in zip(
            old_had_stabilizers + old_idle_stabilizers,
            new_had_stabilizers + new_idle_stabilizers,
            strict=True,
        )
    }
    stab_evolution |= {
        new_wall_stab.uuid: (old_wall_i_stab.uuid, old_wall_h_stab.uuid)
        for old_wall_i_stab, old_wall_h_stab, new_wall_stab in zip(
            old_wall_stabilizers_idle,
            old_wall_stabilizers_had_ordered,
            new_wall_stabilizers,
            strict=True,
        )
    }
    interpretation_step.stabilizer_evolution.update(stab_evolution)

    # C.6) Define stabilizer updates and append to interpretation step
    for stab_new, stab_old_idle in zip(
        new_wall_stabilizers,
        old_wall_stabilizers_idle,
        strict=True,
    ):
        qubits_measured = set(stab_old_idle.data_qubits).intersection(qubits_to_measure)
        # Find cbits from y measurements associated with the stabilizer
        cbits = tuple(
            cbit
            for cbit, qubit_measured in zip(y_meas_cbits, qubits_to_measure)
            if qubit_measured in qubits_measured
        )
        current_updates = interpretation_step.stabilizer_updates.get(stab_new.uuid, ())
        # Add the cbits to the stabilizer updates and add 1 that stems from the
        # Y measurement that joins the 2 stabilizers on each side of the wall
        interpretation_step.stabilizer_updates[stab_new.uuid] = (
            current_updates + cbits + (1,)
        )

    return new_had_stabilizers + new_idle_stabilizers + new_wall_stabilizers


def measure_y_and_hadamard_circuit(
    block: RotatedSurfaceCode,
    qubits_to_measure: list[tuple[int, ...]],
    qubits_to_hadamard: list[tuple[int, ...]],
    interpretation_step: InterpretationStep,
) -> tuple[Circuit, list[Cbit]]:
    """
    Generates a circuit that measures all the data qubits of the wall in the Y basis
    and applies Hadamard gates to the appropriate qubits. It also returns the classical
    bits corresponding to the Y measurements.

    Parameters
    ----------
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    qubits_to_measure : list[tuple[int, ...]]
        The list of qubits to measure.
    qubits_to_hadamard : list[tuple[int, ...]]
        The list of qubits to apply Hadamard gates to.
    interpretation_step : InterpretationStep
        The interpretation step. Note that it may be mutated by generating new
        channels.

    Returns
    -------
    Circuit
        The circuit to create the Y wall.
    list[Cbit]
        The classical bits corresponding to the Y measurements.
    """
    # Get classical channels corresponding to the qubits_to_measure
    cbits = [interpretation_step.get_new_cbit_MUT(f"c_{q}") for q in qubits_to_measure]
    cbit_channels = [
        interpretation_step.get_channel_MUT(
            f"{cbit[0]}_{cbit[1]}", channel_type="classical"
        )
        for cbit in cbits
    ]

    # Get the circuit that measures all the data qubits in the Y basis
    y_wall_measurement_circuit = Circuit(
        name=(f"Measure wall of qubits in the Y basis for block {block.unique_label}"),
        circuit=[
            [
                Circuit(
                    "measure_y",
                    channels=[interpretation_step.get_channel_MUT(qubit), cbit_channel],
                )
                for qubit, cbit_channel in zip(
                    qubits_to_measure, cbit_channels, strict=True
                )
            ]
        ],
    )

    hadamard_circuit = Circuit(
        name=(
            f"transversal hadamard on the data qubits for the y_wall_out operation "
            f"of block {block.unique_label}"
        ),
        circuit=[
            [
                Circuit("h", channels=[interpretation_step.get_channel_MUT(q)])
                for q in qubits_to_hadamard
            ]
        ],
    )

    circuit = Circuit(
        name="y_wall_out - y wall measurement and hadamard",
        circuit=((y_wall_measurement_circuit, hadamard_circuit),),
    )

    return circuit, cbits


def find_new_x_logical_operator(
    block: RotatedSurfaceCode,
    is_top_left_bulk_stab_x: bool,
    qubits_to_idle: list[tuple[int, int, int]],
) -> tuple[PauliOperator, tuple[Stabilizer, ...]]:
    """
    Find the new X logical operator and the stabilizers for the logical operator to jump
    across.

    Parameters
    ----------
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    is_top_left_bulk_stab_x : bool
        Whether the top left bulk stabilizer is X or Z.
    qubits_to_idle : list[tuple[int, int, int]]
        The qubits to idle.

    Returns
    -------
    tuple[PauliOperator, tuple[Stabilizer, ...]]
        The new X logical operator and the stabilizers for the logical operator to jump
        across.
    """
    # C.1) Find where the X logical operator should be depending on the geometry
    # Note that the transversal hadamard operation makes the topological corner jump
    # across the block.
    # So even though for a vertical block with the top left stabilizer as Z, the
    # topological corner is mid-left of the block, after the transversal hadamard
    # operation, it will be mid-right of the block.
    match (block.is_horizontal, is_top_left_bulk_stab_x):
        case (False, False):
            x_log_op_side = Direction.RIGHT
        case (False, True):
            x_log_op_side = Direction.LEFT
        case (True, False):
            x_log_op_side = Direction.BOTTOM
        case (True, True):
            x_log_op_side = Direction.TOP

    # C.2) Create the new X logical operator
    # It should contain all boundary idling qubits on the side where the logical
    # operator has to be placed.
    x_log_qubits = [
        qub for qub in block.boundary_qubits(x_log_op_side) if qub in qubits_to_idle
    ]
    new_x_logical_operator = PauliOperator(
        pauli="X" * len(x_log_qubits),
        data_qubits=x_log_qubits,
    )

    # C.3) Find appropriate stabilizers for the logical operator evolution
    # The stabilizers for the logical operator to jump across requires ALL the
    # stabilizers of the block that contain the qubits to idle.
    stabilizers_for_x_operator_jump = tuple(
        stab
        for stab in block.stabilizers
        if set(stab.data_qubits).intersection(qubits_to_idle)
    )
    return new_x_logical_operator, stabilizers_for_x_operator_jump


def find_qubit_sets(
    block: RotatedSurfaceCode, wall_position: int, is_wall_hor: bool
) -> tuple[list[tuple[int, int, int]], ...]:
    """
    Find the qubits to measure, to idle, and to Hadamard.

    Parameters
    ----------
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    wall_position : int
        The position of the wall.
    is_wall_hor : bool
        Whether the wall is horizontal or vertical.

    Returns
    -------
    tuple[list[tuple[int, int, int]], ...]
        The qubits to measure, to idle, and to Hadamard.
    """

    had_side_unit_vector = (0, 1) if is_wall_hor else (1, 0)
    idle_side_unit_vector = (0, -1) if is_wall_hor else (-1, 0)

    # Find the qubits to measure
    qubits_to_measure = (
        [(q[0], q[1] + wall_position, 0) for q in block.boundary_qubits("top")]
        if is_wall_hor
        else [(q[0] + wall_position, q[1], 0) for q in block.boundary_qubits("left")]
    )
    # Find the qubits to idle (top or left qubits depending on the orientation)
    qubits_to_idle = [
        (q[0] + idle_side_unit_vector[0] * d, q[1] + idle_side_unit_vector[1] * d, q[2])
        for q in qubits_to_measure
        for d in range(1, wall_position + 1)
    ]
    # Find the qubits to Hadamard
    # (bottom or right qubits depending on the orientation)
    qubits_to_hadamard = [
        (
            q[0] + had_side_unit_vector[0] * d,
            q[1] + had_side_unit_vector[1] * d,
            q[2],
        )
        for q in qubits_to_measure
        for d in range(1, max(block.size) - wall_position)
    ]

    return qubits_to_measure, qubits_to_idle, qubits_to_hadamard



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations

from pydantic.dataclasses import dataclass
from pydantic import field_validator, model_validator, Field

from .pauli_operator import PauliOperator
from .utilities.validation_tools import (
    distinct_error,
    dataclass_config,
    ensure_tuple,
    coordinate_length_error,
)


@dataclass(config=dataclass_config)
class Stabilizer(PauliOperator):
    """
    A stabilizer, representing the parity of a set of data qubits in the basis as
    defined in `pauli`.

    Parameters
    ---------
    pauli : str
        The Pauli string of the stabilizer, i.e. in which basis the data qubits are
        included in the stabilizer.

    data_qubits : tuple[tuple[int, ...], ...]
        Data qubits involved in the stabilizer. They are referred to by their
        coordinates in the lattice.

    uuid : str
        Unique identifier of the stabilizer. This is automatically set to a random UUID.

    ancilla_qubits : tuple[tuple[int, ...], ...], optional
        Ancilla qubits involved in the stabilizer. They are referred to by their
        coordinates in the lattice. Default is an empty tuple.
    """

    ancilla_qubits: tuple[tuple[int, ...], ...] = Field(
        default_factory=tuple, validate_default=True
    )

    # Validation functions
    @model_validator(mode="before")
    @classmethod
    def _calculate_ancilla_qubits_from_nr_of_ancillae(cls, data):
        """
        Calculate ancilla qubit indices when the number of ancillae is
        specified. This is useful when deserializing Workbench-generated JSON
        to Eka Block.
        """
        if hasattr(data, "kwargs") and data.kwargs is not None:
            if "nr_of_ancillae" not in data.kwargs or "ancilla_qubits" in data.kwargs:
                return data
        else:
            return data

        nr_of_ancillae = data.kwargs.pop("nr_of_ancillae")
        if nr_of_ancillae == 0:
            return data

        # This is a convention to choose coordinates of ancilla qubits. We
        # take the coordinates to be those of data_qubit[0] and the third
        # index is taken to be the count of the ancilla qubit, starting
        # from 1.
        ancilla_qubits = [
            (
                data.kwargs["data_qubits"][0][0],
                data.kwargs["data_qubits"][0][1],
                ancilla_qubit + 1,
            )
            for ancilla_qubit in range(nr_of_ancillae)
        ]
        data.kwargs["ancilla_qubits"] = tuple(ancilla_qubits)

        return data

    _validate_ancilla_qubits_list = field_validator("ancilla_qubits", mode="before")(
        ensure_tuple
    )
    _validate_distinct_ancilla_qubits = field_validator(
        "ancilla_qubits", mode="before"
    )(distinct_error)
    _validate_coordinate_lengths_ancilla = field_validator(
        "ancilla_qubits", mode="before"
    )(coordinate_length_error)

    # Magic methods
    # def __str__(self) -> str: Method is inherited from PauliOperator

    def __repr__(self) -> str:
        """
        Return a string representation of the stabilizer.
        """
        return f"{self.pauli} {self.data_qubits} {self.ancilla_qubits}"

    def __eq__(self, other: Stabilizer) -> bool:
        """
        Ignore the uuid in the equality check.
        """
        return (
            self.pauli == other.pauli
            and self.data_qubits == other.data_qubits
            and self.ancilla_qubits == other.ancilla_qubits
        )

    def __hash__(self):
        """
        Ignore the uuid in hashing.
        """
        return hash((self.pauli, self.data_qubits, self.ancilla_qubits))



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.interpreter import InterpretationStep
from loom.interpreter.applicator import measureblocksyndromes
from loom.eka.utilities import Direction, Orientation, DiagonalDirection
from loom.eka.operations import MeasureBlockSyndromes, Grow, Shrink
from .utilities import add_vector
from .grow import grow
from .shrink import shrink
from .move_corners import move_corners
from .move_block import move_block
from .logical_phase_via_ywall import move_logical
from ..code_factory import RotatedSurfaceCode
from ..operations import MoveBlock, RotateBlock


def rotate_block(
    interpretation_step: InterpretationStep,
    operation: RotateBlock,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Rotate the boundary of a rotated surface code block by growing it in the specified
    direction, moving the corners in a fault-tolerant manner and shrinking back to the
    original size.

    The rotation is performed by:

    - A.) Begin RotateBlock composite operation session

    - B.) Validity checks
        - B.1) Check that the block is a valid RotatedSurfaceCode block for RotateBlock

    - C.) Grow the block in the specified direction
        - C.1) Grow the block to double its size minus one in the specified direction
        - C.2) Measure syndromes to complete the grow operation
        - C.3) Move the logical operators so that they are correctly located on the
                top-left qubit of the grown block

    - D.) Move the corners appropriately
        - D.1) Move each corner individually
        - D.2) Measure syndromes to complete the corner move in a fault-tolerant manner

    - E.) Shrink the block
        - E.1) Shrink the block from the grown side by distance - 2
        - E.2) Shrink the block from the opposite side by 1 to return to original size

    - F.) Move block
        - F.1) Move the block so that it's occupying the same data qubits as initially
        - F.2) Move the logical operators so that they are correctly located on the
                top-left qubit of the shrunk block

    - G.) Final Circuit
        - G.1) End the composite operation session and append the full rotate block \
        circuit

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The current interpretation step containing the block to be rotated.
    operation : RotateBlock
        The RotateBlock operation specifying the block and direction.
    same_timeslice : bool
        Flag to apply the operation in the same timeslice as the previous operation(s).
    debug_mode: bool
        Flag to apply validation of the new block or not.

    Returns
    -------
    InterpretationStep
        The updated interpretation step after applying the RotateBlock operation.
    """
    # Extract operation parameters
    rsc_block = interpretation_step.get_block(operation.input_block_name)
    grow_direction = operation.grow_direction

    # A.) Begin composite operation session
    interpretation_step.begin_composite_operation_session_MUT(
        same_timeslice=same_timeslice,
        circuit_name=(
            f"Rotate boundary {rsc_block.unique_label} by growing towards "
            f"{grow_direction.name}"
        ),
    )

    # B) Validity checks
    # B.1) Check that the block is a valid RotatedSurfaceCode block for RotateBlock
    rotate_validity_check(rsc_block)

    # Extract useful parameters
    # - distance of the block
    #   Extract the code distance and current block object
    distance = min(rsc_block.size)
    current_block = interpretation_step.get_block(rsc_block.unique_label)

    # C) Grow the block in the specified direction
    # C.1) Grow the block to double its size minus one in the specified direction
    interpretation_step = grow(
        interpretation_step,
        Grow(rsc_block.unique_label, grow_direction, length=distance - 1),
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    current_block = interpretation_step.get_block(rsc_block.unique_label)

    # C.2) Measure syndromes to complete the grow operation
    interpretation_step = measureblocksyndromes(
        interpretation_step,
        MeasureBlockSyndromes(rsc_block.unique_label, n_cycles=distance),
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # C.3) Move the logical operators so that they are correctly located on the top-left
    #     qubit of the grown block
    top_left_qubit = current_block.upper_left_qubit
    for logical_pauli in ["X", "Z"]:
        interpretation_step, current_block = move_logical(
            interpretation_step,
            current_block,
            new_up_left_qubit=top_left_qubit,
            pauli=logical_pauli,
        )

    # D) Move the corners appropriately
    for corner_args in get_boundary_rotation_corner_args_5_1(current_block):
        # D.1) Move each corner individually
        interpretation_step = move_corners(
            interpretation_step,
            current_block,
            corner_args,
            same_timeslice=False,
            debug_mode=debug_mode,
        )
        current_block = interpretation_step.get_block(rsc_block.unique_label)
        # D.2) Measure syndromes to complete the corner move in a fault-tolerant manner
        interpretation_step = measureblocksyndromes(
            interpretation_step,
            MeasureBlockSyndromes(rsc_block.unique_label, n_cycles=distance - 1),
            same_timeslice=False,
            debug_mode=debug_mode,
        )

    # E) Shrink the block
    # E.1) Shrink the block from the grown side by distance - 2
    interpretation_step = shrink(
        interpretation_step,
        Shrink(current_block.unique_label, grow_direction, length=distance - 2),
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    current_block = interpretation_step.get_block(rsc_block.unique_label)

    # E.2) Shrink the block from the opposite side by 1 to return to original size
    interpretation_step = shrink(
        interpretation_step,
        Shrink(current_block.unique_label, grow_direction.opposite(), length=1),
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    current_block = interpretation_step.get_block(rsc_block.unique_label)

    # F) Move block
    # F.1) Move the block so that it's occupying the same data qubits as initially
    interpretation_step = move_block(
        interpretation_step,
        MoveBlock(current_block.unique_label, direction=grow_direction.opposite()),
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    current_block = interpretation_step.get_block(rsc_block.unique_label)

    # F.2) Move the logical operators so that they are correctly located on the top-left
    #     qubit of the shrunk block
    top_left_qubit = current_block.upper_left_qubit
    for logical_pauli in ["X", "Z"]:
        interpretation_step, current_block = move_logical(
            interpretation_step,
            current_block,
            new_up_left_qubit=top_left_qubit,
            pauli=logical_pauli,
        )

    # G) Final Circuit
    # G.1) End the composite operation session and append the full rotate block circuit
    rotate_block_circuit = interpretation_step.end_composite_operation_session_MUT()
    interpretation_step.append_circuit_MUT(rotate_block_circuit, same_timeslice)
    return interpretation_step


def rotate_validity_check(rsc_block: RotatedSurfaceCode):
    """
    Validity checks for applying RotateBlock to a RotatedSurfaceCode block.
    We check that:

    - The block is indeed a RotatedSurfaceCode block
    - The block is square and has odd distance
    - The topological and geometric corners match
    """
    if not isinstance(rsc_block, RotatedSurfaceCode):
        raise ValueError("Cannot apply RotateBlock to non-RotatedSurfaceCode blocks.")
    if not (rsc_block.size[0] == rsc_block.size[1] and rsc_block.size[0] % 2 == 1):
        raise ValueError(
            "Cannot apply RotateBlock to non-square blocks or blocks with even "
            "distance."
        )
    if not set(rsc_block.topological_corners) == set(rsc_block.geometric_corners):
        raise ValueError(
            "Cannot apply RotateBlock to blocks when the topological and geometric "
            "corners do not match."
        )


def get_boundary_rotation_corner_args_5_1(
    rsc_block: RotatedSurfaceCode,
) -> tuple[tuple[tuple[int, ...], Direction, int], ...]:
    """
    Rotate the boundaries by using move_corners 5 times, moving 1 corner at
    a time. The corners are moved in a way such that the distance of the block for all
    logical operators is at least the original distance, ensuring fault-tolerance.
    For this to be possible, the block must be of size (d, 2 * d - 1) or (2 * d - 1, d)
    with odd d.

    Parameters
    ----------
    rsc_block : RotatedSurfaceCode
        The grown rectangular rotated surface code block.

    Returns
    -------
    tuple[tuple[tuple[int, ...], Direction, int], ...]
        A tuple containing 5 tuples, each specifying the arguments for a call to
        move_corners. Each inner tuple contains:

            - The coordinates of the corner to move (as a tuple of integers).
            - The direction in which to move the corner (as a Direction enum).
            - The distance to move the corner (as an integer).
    """
    long_orientation = (
        Orientation.HORIZONTAL
        if rsc_block.size[0] > rsc_block.size[1]
        else Orientation.VERTICAL
    )

    # Find the corner qubits for ease of access
    tl = rsc_block.get_corner_from_direction(
        DiagonalDirection.from_directions((Direction.TOP, Direction.LEFT))
    )
    tr = rsc_block.get_corner_from_direction(
        DiagonalDirection.from_directions((Direction.TOP, Direction.RIGHT))
    )
    br = rsc_block.get_corner_from_direction(
        DiagonalDirection.from_directions((Direction.BOTTOM, Direction.RIGHT))
    )
    bl = rsc_block.get_corner_from_direction(
        DiagonalDirection.from_directions((Direction.BOTTOM, Direction.LEFT))
    )

    # Get the distances
    short_distance = min(rsc_block.size)
    long_distance = max(rsc_block.size)

    if long_distance != 2 * short_distance - 1 or short_distance % 2 != 1:
        raise ValueError(
            "Cannot obtain corner moves for blocks that are not of size (d, 2 * d -1)"
            f" or (2 * d - 1, d) with odd d. Given size: {rsc_block.size}"
        )

    # The lengths to move each corner
    short_length = min(rsc_block.size) - 1
    long_length = max(rsc_block.size) - 1

    # Return the appropriate sequence of corner moves
    match (rsc_block.weight_2_stab_is_first_row, long_orientation):
        case (False, Orientation.VERTICAL):
            return (
                ((tl, Direction.BOTTOM, short_length),),  # Move TL halfway down
                ((tr, Direction.LEFT, short_length),),  # Move TR left
                ((br, Direction.TOP, long_length),),  # Move BR up
                ((bl, Direction.RIGHT, short_length),),  # Move BL right
                (
                    (
                        add_vector(tl, Direction.BOTTOM, short_length),
                        Direction.BOTTOM,
                        short_length,
                    ),
                ),  # Move initial corner all the way down
            )
        case (True, Orientation.VERTICAL):
            return (
                ((tr, Direction.BOTTOM, short_length),),  # Move TR halfway down
                ((tl, Direction.RIGHT, short_length),),  # Move TL right
                ((bl, Direction.TOP, long_length),),  # Move BL up
                ((br, Direction.LEFT, short_length),),  # Move BR left
                (
                    (
                        add_vector(tr, Direction.BOTTOM, short_length),
                        Direction.BOTTOM,
                        short_length,
                    ),
                ),  # Move initial corner all the way down
            )
        case (False, Orientation.HORIZONTAL):
            return (
                ((bl, Direction.RIGHT, short_length),),  # Move BL halfway right
                ((tl, Direction.BOTTOM, short_length),),  # Move TL down
                ((tr, Direction.LEFT, long_length),),  # Move TR left
                ((br, Direction.TOP, short_length),),  # Move BR up
                (
                    (
                        add_vector(bl, Direction.RIGHT, short_length),
                        Direction.RIGHT,
                        short_length,
                    ),
                ),  # Move initial corner all the way right
            )
        case (True, Orientation.HORIZONTAL):
            return (
                ((tl, Direction.RIGHT, short_length),),  # Move TL halfway right
                ((bl, Direction.TOP, short_length),),  # Move BL up
                ((br, Direction.LEFT, long_length),),  # Move BR left
                ((tr, Direction.BOTTOM, short_length),),  # Move TR down
                (
                    (
                        add_vector(tl, Direction.RIGHT, short_length),
                        Direction.RIGHT,
                        short_length,
                    ),
                ),  # Move initial corner all the way right
            )
        case _:
            raise ValueError(
                "Invalid combination of weight_2_stab_is_first_row and "
                "long_orientation"
            )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import numpy as np


def paulichar_to_xz(p: str) -> tuple[int, int]:
    """
    Function that turns a Pauli into a pair of x,z bits.

    Parameters
    ----------
    p : str
        The Pauli character.

    Returns
    -------
    tuple[int, int]
        x, z bits

    Raises
    ------
    ValueError
        If the Pauli is not I, Z, X or Y.
    """

    match p:
        case "_" | "I" | "i":
            return (0, 0)
        case "Z" | "z":
            return (0, 1)
        case "X" | "x":
            return (1, 0)
        case "Y" | "y":
            return (1, 1)
        case _:
            raise ValueError(
                "The Pauli character should be I, Z, X or Y or their "
                "lower case versions. _ is also accepted as I."
            )


def paulichar_to_xz_npfunc(p: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Vectorized version of paulichar_to_xz.

    Parameters
    ----------
    p : np.ndarray
        An array of Pauli characters.

    Returns
    -------
    tuple[np.ndarray]
        Two arrays of x and z bits.
    """

    return np.frompyfunc(paulichar_to_xz, 1, 2)(p)


def paulixz_to_char(
    x: int,
    z: int,
) -> str:
    """
    Function that turns a pair of x, z bits into their Pauli as a str.

    Parameters
    ----------
    x : int
        The x bit.
    z : int
        The z bit.

    Returns
    -------
    str
        The Pauli character as a string.

    Raises
    ------
    ValueError
        If x or z are not 0 or 1.
    """

    match (x, z):
        case (0, 0):
            return "_"
        case (0, 1):
            return "Z"
        case (1, 0):
            return "X"
        case (1, 1):
            return "Y"
        case _:
            raise ValueError("The x and z values should be 0 or 1.")


def paulixz_to_char_npfunc(x: np.ndarray, z: np.ndarray) -> np.ndarray:
    """
    Vectorized version of paulixz_to_char.

    Parameters
    ----------
    x : np.ndarray
        An array of x bits.
    z : np.ndarray
        An array of z bits.

    Returns
    -------
    np.ndarray
        An array of Pauli characters.
    """

    return np.frompyfunc(paulixz_to_char, 2, 1)(x, z)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import numpy as np
from pydantic import Field
from pydantic.dataclasses import dataclass
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from loom.eka import Stabilizer, Lattice, Block, PauliOperator

from .plotting_utils import (
    order_points_counterclockwise,
    hex_to_rgb,
    draw_half_circle,
    center_of_scatter_plot,
    get_font_color,
    average_color_hex,
)


@dataclass()
class StabilizerPlot:  # pylint: disable=too-many-instance-attributes
    """
    Class for plotting stabilizers, pauli strings, and individual qubits of a Eka.

    Parameters
    ----------
    lattice : Lattice
        Lattice for which the plot should be created.
    stabilizers : list[Stabilizer] | None
        List of stabilizers which should be plotted
    width : int | None
        Width of the plotly figure. Default is 800.
    height : int | None
        Height of the plotly figure. Default is 600.
    rescale_coords : bool | None
        In case the image is drawn with a canvas for stabilizers containing different
        pauli operators, this parameter has to be set to True. Default is False. When
        True, the mixed-type stabilizers are drawn pixel by pixel for a canvas of the
        given width and height. The coordinates of the data qubits are rescaled to fit
        the image.
    padding_relative : float | None
        In case the image is drawn with a canvas for stabilizers containing different
        pauli operators, this parameter specifies the padding around the drawings,
        relative to the width of the image.
    xmin : float | None
        Minimum x-coordinate of the data qubit locations. Only needed if
        `rescale_coords` is True. Since objects might be added to the figure later,
        the x and y ranges have to be defined already at the beginning.
    xmax : float | None
        Maximum x-coordinate of the data qubit locations. Only needed if
        `rescale_coords` is True. Since objects might be added to the figure later,
        the x and y ranges have to be defined already at the beginning.
    ymin : float | None
        Minimum y-coordinate of the data qubit locations. Only needed if
        `rescale_coords` is True. Since objects might be added to the figure later,
        the x and y ranges have to be defined already at the beginning.
    ymax : float | None
        Maximum y-coordinate of the data qubit locations. Only needed if
        `rescale_coords` is True. Since objects might be added to the figure later,
        the x and y ranges have to be defined already at the beginning.
    title : str | None
        Title of the figure.
    opacity_stabs : float | None
        Opacity for the stabilizer colors. Range: 0 to 1 with 0 being
        completely transparent and 1 being completely opaque.
    show_grid : bool | None
        If set to False (default), the x and y axes are not shown, the zero line is
        hidden, and the grid is not displayed.
    fill_colors : dict[str, str] | None
        Dict for the fill colors of X, Y, and Z stabilizers.
    line_colors : dict[str, str] | None
        Dict for the line colors of X, Y, and Z stabilizers.
    dqb_plot_indices : bool | None
        If set to True (default), the data qubit indices are plotted inside the markers.
    dqb_marker_size : int | list[int] | None
        Marker size for the data qubits. Either a single integer is provided which is
        used for all data qubits, or a list of integers is provided which specifies the
        marker size for each data qubit individually.
    dqb_marker_color : str | list[str] | None
        Color for the data qubit markers. Either a single color is provided which is
        used for all data qubits, or a list of colors is provided which specifies the
        marker color for each data qubit individually.
    """

    lattice: Lattice
    stabilizers: list[Stabilizer] | None = Field(default=None, validate_default=True)
    width: int | None = 800
    height: int | None = 600
    rescale_coords: bool | None = False
    padding_relative: float | None = 0.05
    xmin: float | None = None
    xmax: float | None = None
    ymin: float | None = None
    ymax: float | None = None
    title: str | None = None
    opacity_stabs: float | None = 0.8
    show_grid: bool | None = False
    fill_colors: dict[str, str] | None = Field(
        default_factory=lambda: {
            "X": "#f2a04c",
            "Y": "#e95cb5",
            "Z": "#49bbc2",
        }
    )
    line_colors: dict[str, str] | None = Field(
        default_factory=lambda: {
            "X": "#7f6668",
            "Y": "#7f6668",
            "Z": "#7f6668",
        }
    )
    dqb_plot_indices: bool | None = True
    dqb_marker_size: int | list[int] | None = 25
    dqb_marker_color: str | list[str] | None = "#ec6f6d"

    def __post_init__(self) -> None:
        if self.rescale_coords:
            if None in [self.xmin, self.xmax, self.ymin, self.ymax]:
                raise ValueError(
                    "If `rescale_coords` is set to True, `xmin`, `xmax`, `ymin`, and "
                    "`ymax` must be provided."
                )

        # Create the figure object
        self._fig = make_subplots(rows=1, cols=1)
        self.canvas = np.ones((self.height, self.width, 3), dtype=np.uint8)
        self.canvas[:, :, 0] = 229
        self.canvas[:, :, 1] = 236
        self.canvas[:, :, 2] = 246

        # Calculate coordinates of data qubits and store them both in a list and a dict
        self.dqubit_coords = [self.get_qb_coord(qb) for qb in self.lattice.all_qubits()]
        self.dqubit_coordinates_map = {
            qb: self.get_qb_coord(qb) for qb in self.lattice.all_qubits()
        }

        # Make sure dqb_marker_color is a list of colors
        if isinstance(self.dqb_marker_color, str):
            self.dqb_marker_color = [self.dqb_marker_color] * len(self.dqubit_coords)
        else:
            if len(self.dqb_marker_color) != len(self.dqubit_coords):
                raise ValueError(
                    "The length of the dqb_marker_color list should be equal "
                    "to the number of data qubits. The list has a "
                    f"length of {len(self.dqb_marker_color)} but there are "
                    f"{len(self.dqubit_coords)} data qubits."
                )

        # Color of data qubit markers in the legend
        self.legend_dqb_marker_color = "#000000"

        # Dummy plot of data qubits such that they appear first in the legend
        self._fig.add_trace(
            go.Scatter(
                x=[None],
                y=[None],
                mode="markers",
                name="Data qubits",
                marker={
                    "size": self.dqb_marker_size,
                    "color": self.dqb_marker_color[0],
                    "line": {
                        "width": 1,
                        "color": "#9f4b49",
                    },
                },
                legendgroup="data_qubits",
                showlegend=True,
            )
        )

        # Settings
        self._fig.update_layout(
            width=self.width,
            height=self.height,
            xaxis_title="X-axis",
            yaxis_title="Y-axis",
            showlegend=True,
            xaxis={
                "showgrid": self.show_grid,
                "zeroline": False,
                "visible": self.show_grid,
            },
            yaxis={
                "showgrid": self.show_grid,
                "visible": self.show_grid,
                "scaleanchor": "x",
                "scaleratio": 1,
                "autorange": "reversed",  # Otherwise y axis would be reversed
            },
            hoverdistance=1,  # With this setting, the hover text will only show up when
            # exactly hovering above an object and not when one is close to it. This is
            # important for overlapping objects where the wrong hover text might be
            # shown.
        )
        if self.title is not None:
            self._fig.update_layout(title=self.title)

    def _scale_coords(self, x: float, y: float) -> tuple[float, float]:
        """
        If `self.rescale_coords == True`, any coordinates are rescaled to fit the
        canvas dimensions. This is used when drawing stabilizers with different pauli
        operators with a color gradient pixel by pixel.

        NOTE: In most cases, `self.rescale_coords == False` and this functions does
        nothing.

        Parameters
        ----------
        x : float
            x coordinate of the pixel to be rescaled
        y : float
            y coordinate of the pixel to be rescaled

        Returns
        -------
        tuple[float, float]
            x and y coordinate of the rescaled point
        """
        if self.rescale_coords:
            # For a good visualization, we do not want to plot objects exactly until the
            # border of the canvas but keep some distance (padding) to the border.
            # By default, this padding is 5% of the width of the canvas.
            padding_absolute = self.padding_relative * self.width
            # New x coordinate
            # Previously values were in the range [xmin, xmax] and now we want to scale
            # them to the range [padding_absolute, width - padding_absolute].
            # This corresponds to scaling by a factor of
            # (width - 2 * padding_absolute) / (xmax - xmin) and then shifting by
            # padding_absolute.
            x_new = padding_absolute + (x - self.xmin) * (
                self.width - 2 * padding_absolute
            ) / (self.xmax - self.xmin)
            # New y coordinate
            # Calculation is analogous as for x
            y_new = padding_absolute + (y - self.ymin) * (
                self.height - 2 * padding_absolute
            ) / (self.ymax - self.ymin)
            return x_new, y_new

        return x, y

    def show(self, **kwargs) -> None:
        """Show the plotly figure."""
        self._fig.show(**kwargs)

    def get_qb_coord(self, qb: tuple[int, ...]) -> tuple[float, ...]:
        """
        Get the coordinates of a data qubit. The coordinates of data qubit (x, y, a)
        are calculated by the formula x * l[0] + y * l[1] + a * b[a], where l is the
        list of lattice vectors and b is the list of basis vectors. We put all these
        terms into a list and sum them up in the very end.
        """
        list_vectors = [
            np.array(self.lattice.lattice_vectors[i]) * qb[i]
            for i in range(len(self.lattice.lattice_vectors))
        ]
        # If there are more entries in the qubit index than lattice vectors, the last
        # element in the tuple denotes which basis vector in the unit cell the qubit
        # belongs to
        if len(qb) > len(self.lattice.lattice_vectors):
            list_vectors.append(self.lattice.basis_vectors[qb[-1]])
        else:
            # If there is no such entry, the qubit is assumed to belong to the first
            # basis vector in the unit cell
            list_vectors.append(self.lattice.basis_vectors[0])
        return np.sum(np.array(list_vectors), axis=0)

    # pylint: disable=too-many-locals
    def get_stabilizer_traces(
        self,
        stabilizers: list[Stabilizer],
        fill_colors: dict[str, str] | None = None,
        fill: str = "toself",
        opacity_stabs: float | None = None,
        showlegend: bool = True,
        legendgroup: str | None = None,
    ) -> list[go.Scatter]:
        """
        Generate the plotly traces for the stabilizers.

        Parameters
        ----------
        stabilizers : list[Stabilizer]
            Stabilizers which should be plotted.
        fill_colors : dict[str, str] | None
            Colors for the fill of the stabilizers. If None is provided, the default
            colors from the class are taken.
        fill : str
            Defines how the area of the stabilizers is filled. Default is 'toself' which
            fills the area of the polygon. Another option is 'none' (no fill) for having
            only the border line but no filling.
        opacity_stabs : float | None
            Opacity for the stabilizer colors. If If None is provided, the default
            opacity from the class is taken.
        showlegend : bool, optional
            If True (the default), the stabilizers are shown in the legend.
        legendgroup: str | None
            Legend group to which the stabilizers belong.

        Returns
        -------
        list[go.Scatter]
            List of scatter plot traces for the stabilizers.
        """
        stab_traces = []
        for stab in stabilizers:
            polygon_corners = [
                self.dqubit_coordinates_map[qb] for qb in stab.data_qubits
            ]

            # Order the corners of the polygon such that plotting the polygon looks nice
            ordered_corners = order_points_counterclockwise(polygon_corners)
            ordered_corners_scaled = []
            for corner in ordered_corners:
                corner_x_scaled, corner_y_scaled = self._scale_coords(
                    corner[0], corner[1]
                )
                ordered_corners_scaled.append([corner_x_scaled, corner_y_scaled])

            stab_name = ""
            for pauli, data_qubit in zip(stab.pauli, stab.data_qubits, strict=True):
                stab_name += pauli + "<sub>" + str(data_qubit) + "</sub>"

            # Line color
            line_color = average_color_hex(
                [self.line_colors[pauli] for pauli in stab.pauli]
            )

            # Start with self.fill_colors and update it with values provided in the
            # fill_colors argument of this function
            stab_fill_colors = self.fill_colors
            if fill_colors is not None:
                stab_fill_colors.update(fill_colors)

            if stab.pauli in stab_fill_colors:
                fill_color = stab_fill_colors[stab.pauli]
            else:
                fill_color = average_color_hex(
                    [stab_fill_colors[pauli] for pauli in stab.pauli]
                )

            if opacity_stabs is None:
                opacity_stabs = self.opacity_stabs
            background_color_arr = hex_to_rgb(fill_color)
            background_color = "rgba("
            for color in background_color_arr:
                background_color += str(color) + ","
            background_color = background_color + str(opacity_stabs) + ")"

            if (
                len(stab.data_qubits) == 2
            ):  # Weight-2 stabilizer --> Plot stabilizer as half circle
                (x1, y1) = self.dqubit_coordinates_map[stab.data_qubits[0]]
                (x2, y2) = self.dqubit_coordinates_map[stab.data_qubits[1]]
                x1_scaled, y1_scaled = self._scale_coords(x1, y1)
                x2_scaled, y2_scaled = self._scale_coords(x2, y2)
                dist = np.linalg.norm([x2_scaled - x1_scaled, y2_scaled - y1_scaled])
                direction = np.arctan2(y2 - y1, x2 - x1)

                # Generate two half circles for the two possible directions in which
                # it could point. Then check which of the two is better.
                # For codes with dimension > 1, the half circles will face away
                # from the bulk of the code block.
                # In the special case of repetition codes, the half circles will
                # be turned into different directions alternatingly.
                xscaled, yscaled = self._scale_coords((x1 + x2) / 2, (y1 + y2) / 2)
                scatter1 = draw_half_circle(
                    [xscaled, yscaled],
                    dist / 2,
                    direction,
                    name=stab_name,
                    fillcolor=background_color,
                    line={"color": line_color},
                    text=stab_name,
                    showlegend=showlegend,
                    legendgroup=legendgroup,
                )
                scatter2 = draw_half_circle(
                    [xscaled, yscaled],
                    dist / 2,
                    direction + np.pi,
                    name=stab_name,
                    fillcolor=background_color,
                    line={"color": line_color},
                    text=stab_name,
                    showlegend=showlegend,
                    legendgroup=legendgroup,
                )
                center = [q for stab in stabilizers for q in stab.data_qubits]
                center_of_block = np.average(np.array(center), axis=0)
                scaled_center = self._scale_coords(
                    center_of_block[0], center_of_block[1]
                )
                center1 = center_of_scatter_plot(scatter1)
                center2 = center_of_scatter_plot(scatter2)
                dist1 = np.linalg.norm(
                    [scaled_center[0] - center1[0], scaled_center[1] - center1[1]]
                )
                dist2 = np.linalg.norm(
                    [scaled_center[0] - center2[0], scaled_center[1] - center2[1]]
                )
                if dist1 < dist2:
                    stab_traces.append(scatter2)
                else:
                    stab_traces.append(scatter1)
            else:  # Weight > 2 --> Plot stabilizer as polygon
                ordered_corners_scaled.append(ordered_corners_scaled[0])
                ordered_corners_scaled_x = [
                    coord[0] for coord in ordered_corners_scaled
                ]
                ordered_corners_scaled_y = [
                    coord[1] for coord in ordered_corners_scaled
                ]

                stab_traces.append(
                    go.Scatter(
                        x=ordered_corners_scaled_x,
                        y=ordered_corners_scaled_y,
                        mode="lines",
                        name=stab_name,
                        fill=fill,  # Fill the area of the polygon
                        fillcolor=background_color,
                        line={"color": line_color},
                        text=stab_name,
                        hoverinfo="text",
                        showlegend=showlegend,
                        legendgroup=legendgroup,
                    )
                )

        return stab_traces

    def add_stabilizers(
        self,
        stabilizers: list[Stabilizer],
        **kwargs,
    ) -> None:
        """
        Adds stabilizers to the plot.

        Parameters
        ----------
        stabilizers : list[Stabilizer]
            Stabilizers which should be plotted.
        """
        stabilizers_traces = self.get_stabilizer_traces(stabilizers, **kwargs)

        for stab in stabilizers_traces:
            self._fig.add_trace(stab)

    # pylint: disable=too-many-branches
    def get_dqubit_traces(
        self,
        dqb_show: list[bool] | None = None,
        labels: list[str] | None = None,
        marker_mode: str | None = None,
        marker_style: dict | None = None,
        legendgroup: str | None = None,
        marker_opacity: float | None = 1,
    ) -> list[go.Scatter]:
        """
        Generate the plotly traces for the data qubit markers.

        Parameters
        ----------
        dqb_show : list[bool] | None
            List of booleans, indicating whether the i-th data qubit should be
            shown or not. If None, all data qubits are shown.
        labels : list[str] | None
            Labels which should be displayed inside the data qubit markers or on hover.
        marker_mode : str | None
            Mode in which the markers are displayed, e.g. `markers` or `markers+text`
        marker_style : dict | None
            Define the style of the markers such as size, color, shape, etc.
        legendgroup : str | None
            Legend group to which the data qubit markers belong.
        marker_opacity : float | None
            Opacity of the data qubit markers.

        Returns
        -------
        list[go.Scatter]
            List of scatter plot traces for the data qubit markers.
        """
        # Check inputs and transform them
        if dqb_show is None:
            dqb_show = [True] * len(self.dqubit_coordinates_map)  # Plot all data qubits
        else:
            if len(dqb_show) != len(self.dqubit_coordinates_map):
                raise ValueError(
                    "The length of the dqb_show list should be equal "
                    "to the number of data qubits. The list has a "
                    f"length of {len(dqb_show)} but there are "
                    f"{len(self.dqubit_coordinates_map)} data qubits."
                )

        # Legend group
        if legendgroup is None:
            legendgroup = "data_qubits"

        # Marker type
        if marker_mode is None:
            if self.dqb_plot_indices:
                marker_mode = "markers+text"
            else:
                marker_mode = "markers"

        dqubit_traces = []
        # Iterate over all data qubits and (if dqb_show[i] is True for the data qubit
        # with index i) create a trace for this data qubit marker.
        # We create separate traces for the data qubits to be able to change the style
        # and text of each data qubit individually.
        for i, dqubit_coord in enumerate(self.dqubit_coords):
            # Check if this data qubit should be plotted or not
            if dqb_show[i] is not True:
                continue
            # Label
            if labels is not None:
                text = labels[i]
            else:
                converted_coords = tuple(
                    (
                        coord.item()
                        if isinstance(coord, (np.integer, np.floating))
                        else coord
                    )
                    for coord in list(self.dqubit_coordinates_map.keys())[i]
                )
                if self.dqb_plot_indices:
                    text = str(converted_coords)
                else:
                    text = f"Data qubit {converted_coords}"

            # Marker style
            marker_color_hex = self.dqb_marker_color[i]
            marker_color = "rgba("
            for c in hex_to_rgb(marker_color_hex):
                marker_color += str(c) + ","
            marker_color += str(round(marker_opacity, 3)) + ")"

            if marker_style is None:
                dqb_marker_style = {
                    "size": self.dqb_marker_size,
                    "color": marker_color,
                    "line": {
                        "width": 1,
                        "color": "#9f4b49",
                    },
                }
                textfont_color = get_font_color(marker_color_hex)
            else:
                dqb_marker_style = marker_style
                if "color" not in dqb_marker_style:
                    raise ValueError(
                        "The marker style dict must contain the color of the markers."
                    )
                if len(dqb_marker_style["color"]) == 7:  # Already HEX format
                    textfont_color = get_font_color(dqb_marker_style["color"])
                else:
                    raise ValueError("Only HEX values supported at the moment.")

            xnew, ynew = self._scale_coords(dqubit_coord[0], dqubit_coord[1])
            dqubit_traces.append(
                go.Scatter(
                    x=[xnew],
                    y=[ynew],
                    mode=marker_mode,
                    marker=dqb_marker_style,
                    text=text,
                    textfont_color=textfont_color,
                    textposition="middle center",
                    hoverinfo="text",
                    legendgroup=legendgroup,
                    showlegend=False,
                )
            )
        return dqubit_traces

    def add_dqubit_traces(
        self,
        dqb_show: list[bool] | None = None,
        **kwargs,
    ) -> None:
        """
        Add data qubit markers to the figure.

        Parameters
        ----------
        dqb_show : list[bool] | None
            List of booleans, indicating whether the i-th data qubit should be
            shown or not. If None, all data qubits are shown.
        """
        dqubit_traces = self.get_dqubit_traces(dqb_show, **kwargs)

        for dqb_trace in dqubit_traces:
            self._fig.add_trace(dqb_trace)

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def get_pauli_string_traces(
        self,
        operation: PauliOperator,
        dqb_marker_size: int | None = 25,
        connecting_line: bool | None = True,
        log_op_fill_colors: dict[str, str] | None = None,
        log_op_line_colors: dict[str, str] | None = None,
        legend_name: str | None = "",
        legendgroup: str = "data_qubits_log_op",
        shift: tuple[float, float] = (0, 0),
    ) -> list[go.Scatter]:
        """
        Generate the plotly traces for visualizing a Pauli string.

        Parameters
        ----------
        operation : PauliOperator
            Pauli string which should be plotted
        dqb_marker_size : int | None
            Marker size for the data qubits contained in the Pauli string
        connecting_line : bool | None
            If True, the data qubits in the Pauli string will be connected by a line
        log_op_fill_colors: dict[str, str] | None
            Dict containing the colors for the X, Y, and Z markers of the logical
            operators. If None is provided, the same colors are taken as for the
            stabilizers.
        log_op_line_colors : dict[str, str] | None
            Dict containing the border colors for the X, Y, and Z markers of the logical
            operators. If None is provided, the same colors are taken as for the
            stabilizers.
        legend_name : str | None
            Name of this Pauli string which should appear in the legend.
        legendgroup : str, optional
            Legend group to which the markers belong
        shift : tuple[float, float]
            2D vector by which the data qubit markers should be shifted, relative to the
            original position of the data qubits.

        Returns
        -------
        list[go.Scatter]
            List of scatter plot traces for the data qubit markers contained in the
            pauli string as well as (if specified) a connecting line between them.
        """
        if log_op_fill_colors is None:
            log_op_fill_colors = self.fill_colors
        if log_op_line_colors is None:
            log_op_line_colors = self.line_colors

        # List of coordinates of the data qubits that are involved in the Pauli string
        pauli_dqubit_coords = np.array(
            [
                np.array(self.dqubit_coordinates_map[qb]) + shift
                for qb in operation.data_qubits
            ]
        )

        pauli_string_traces = []

        # Connecting line
        if connecting_line is True:
            connecting_line_color = average_color_hex(
                [log_op_fill_colors[pauli] for pauli in operation.pauli]
            )
            pauli_string_traces.append(
                go.Scatter(
                    x=pauli_dqubit_coords[:, 0],
                    y=pauli_dqubit_coords[:, 1],
                    mode="lines",
                    hoverinfo="text",
                    line={"color": connecting_line_color, "width": 15},
                    showlegend=False,
                    legendgroup=legendgroup,
                )
            )

        # Legend name for the pauli operator
        if legend_name == "":
            for i, qb in enumerate(operation.data_qubits):
                legend_name += operation.pauli[i] + "<sub>" + str(qb) + "</sub>"

        # Create traces for the data qubit markers containing the pauli operator
        for i, data_qubit in enumerate(operation.data_qubits):
            label_pauli_with_number = f"{operation.pauli[i]}{data_qubit}"

            pauli_string_traces.append(
                go.Scatter(
                    x=[pauli_dqubit_coords[i, 0]],
                    y=[pauli_dqubit_coords[i, 1]],
                    mode="markers+text",
                    name=legend_name,
                    marker={
                        "size": dqb_marker_size * 1.25,
                        "color": log_op_fill_colors[operation.pauli[i]],
                        "line": {
                            "width": 1,
                            "color": log_op_line_colors[operation.pauli[i]],
                        },
                        "symbol": "square",
                    },
                    text=f"{operation.pauli[i]}",
                    textposition="middle center",
                    hoverinfo="text",
                    hovertemplate=label_pauli_with_number,
                    legendgroup=legendgroup,
                    showlegend=False,
                )
            )

        return pauli_string_traces

    def plot_pauli_string(
        self,
        operation: PauliOperator,
        **kwargs,
    ) -> None:
        """
        Add a Pauli string plotting to the figure.

        Parameters
        ----------
        operation : PauliOperator
            Pauli string which should be plotted
        """
        for trace in self.get_pauli_string_traces(operation, **kwargs):
            self._fig.add_trace(trace)

    def plot_pauli_charges(
        self,
        block: Block,
        legendgroup: str | None = None,
    ) -> None:
        """
        Plot the Pauli charges of the data qubits of the given block.

        Parameters
        ----------
        block: Block
            Block for which the pauli charges should be plotted
        legendgroup : str | None
            Legend group to which the markers belong
        """
        # Construct a PauliOperator containing the charge of each data qubit
        # respectively
        op = PauliOperator(
            pauli="".join(
                [charge for charge in block.pauli_charges.values() if charge != "_"]
            ),
            data_qubits=[
                qb for qb, charge in block.pauli_charges.items() if charge != "_"
            ],
        )

        pauli_traces = self.get_pauli_string_traces(
            op,
            connecting_line=False,
            legend_name="Pauli charges",
            legendgroup=legendgroup,
        )

        for tr in pauli_traces:
            self._fig.add_trace(tr)

    def plot_blocks(
        self,
        blocks: Block | list[Block],
        plot_logical_operators: bool = True,
        plot_pauli_charges: bool = False,
        log_ops_shift: float = 0.03,
    ) -> None:
        """
        Plot all the stabilizers of a `Block` or multiple `Block` objects.

        Parameters
        ----------
        blocks : Block | list[Block]
            A block or a list of blocks to be plotted
        plot_logical_operators : bool
            Whether the logical operators should be plotted, defaults to True
        plot_pauli_charges : bool
            Whether the pauli charges should be plotted, defaults to False
        log_ops_shift: float = 0.03
            Amount by which the logical X operator is shifted to the bottom left and the
            logical Z operator to the top right. This is done such that one can see both
            boxes at the qubit(s) on which both operators act. Defaults to 0.03.
        """
        if isinstance(blocks, Block):
            blocks = [blocks]

        for block in blocks:
            legendgroup = f"block_{block.unique_label}"

            # Dummy plot for the block to have a nice label in the legend
            self._fig.add_trace(
                go.Scatter(
                    x=[None],
                    y=[None],
                    mode="markers",
                    name=block.unique_label,
                    marker={
                        "size": self.dqb_marker_size,
                        "color": "#555555",
                        "line": {
                            "width": 1,
                            "color": "#333333",
                        },
                    },
                    legendgroup=legendgroup,
                    showlegend=True,
                )
            )

            # Stabilizers
            self.add_stabilizers(
                block.stabilizers, legendgroup=legendgroup, showlegend=False
            )

            # Pauli charges
            if plot_pauli_charges:
                self.plot_pauli_charges(block, legendgroup=legendgroup)

            # Logical X and Z operators
            if plot_logical_operators:
                for log_x in block.logical_x_operators:
                    self.plot_pauli_string(
                        log_x,
                        shift=(-log_ops_shift, -log_ops_shift),
                        legendgroup=legendgroup,
                    )
                for log_z in block.logical_z_operators:
                    self.plot_pauli_string(
                        log_z,
                        shift=(log_ops_shift, log_ops_shift),
                        legendgroup=legendgroup,
                    )
